from __future__ import annotations

import cProfile
import json
import math
import pstats
import statistics
import time
from collections.abc import Callable
from pathlib import Path
from typing import Literal, Protocol, cast

import msgspec
from tqdm import tqdm

from grim.config import CrimsonConfig
from grim.console import ConsoleState
from grim.raylib_api import rl
from grim.view import ViewContext

from ...modes.replay_playback_mode import ReplayPlaybackMode
from ...replay import Replay
from .playback_driver import PlaybackWalkHooks, build_verify_playback_driver
from .render_telemetry import RenderTelemetryFrameSnapshot, RenderTelemetrySession
from .render_telemetry_charts import write_render_telemetry_charts
from .setup import RunResult

ProfileSortKey = Literal["cumtime", "tottime"]
HotspotSource = Literal["project", "all"]


class _ProgressBarLike(Protocol):
    def update(self, value: int) -> None: ...

    def set_postfix_str(self, s: str, refresh: bool = True) -> None: ...

    def close(self) -> None: ...


class _ProfileStatsLike(Protocol):
    stats: dict[tuple[str, int, str], tuple[int, int, float, float, object]]


class ReplayBenchmarkError(ValueError):
    pass


class BenchmarkSample(msgspec.Struct, frozen=True):
    wall_ms: float
    ticks_per_second: float
    realtime_x: float


class BenchmarkAggregate(msgspec.Struct, frozen=True):
    min: float
    p50: float
    mean: float
    p95: float
    max: float
    stdev: float


class ReplayProfileHotspot(msgspec.Struct, frozen=True):
    file: str
    line: int
    function: str
    primitive_calls: int
    total_calls: int
    tottime: float
    cumtime: float


class ReplayProfileResult(msgspec.Struct, frozen=True):
    sort: ProfileSortKey
    top: int
    source: HotspotSource
    hotspots: tuple[ReplayProfileHotspot, ...]


class ReplayRenderTelemetryTopTick(msgspec.Struct, frozen=True):
    tick_index: int
    frame_index: int
    value: float


class ReplayRenderTelemetryFrame(msgspec.Struct, frozen=True):
    frame_index: int
    tick_index_before_update: int
    tick_index_after_update: int
    update_ms: float
    draw_ms: float
    frame_ms: float
    draw_calls_total: int
    draw_calls_by_api: dict[str, int]
    draw_calls_by_pass: dict[str, int]
    pass_ms: dict[str, float]


class ReplayRenderTelemetrySummary(msgspec.Struct, frozen=True):
    frame_ms: BenchmarkAggregate
    update_ms: BenchmarkAggregate
    draw_ms: BenchmarkAggregate
    draw_calls_total: BenchmarkAggregate
    top_draw_ms_ticks: tuple[ReplayRenderTelemetryTopTick, ...]
    top_frame_ms_ticks: tuple[ReplayRenderTelemetryTopTick, ...]
    top_draw_calls_ticks: tuple[ReplayRenderTelemetryTopTick, ...]


class ReplayRenderTelemetryArtifacts(msgspec.Struct, frozen=True):
    telemetry_json_path: str | None = None
    charts_dir: str | None = None
    frame_timing_svg: str | None = None
    draw_calls_svg: str | None = None
    pass_timing_stacked_svg: str | None = None
    report_md: str | None = None


class ReplayRenderTelemetryResult(msgspec.Struct, frozen=True):
    frames: tuple[ReplayRenderTelemetryFrame, ...]
    summary: ReplayRenderTelemetrySummary
    artifacts: ReplayRenderTelemetryArtifacts | None = None
    preview: tuple[ReplayRenderTelemetryFrame, ...] = ()


class ReplayBenchmarkResult(msgspec.Struct, frozen=True):
    run_result: RunResult
    samples: tuple[BenchmarkSample, ...]
    wall_ms: BenchmarkAggregate
    ticks_per_second: BenchmarkAggregate
    realtime_x: BenchmarkAggregate
    profile: ReplayProfileResult | None
    render_telemetry: ReplayRenderTelemetryResult | None = None


class _RenderOnceResult(msgspec.Struct, frozen=True):
    run_result: RunResult
    telemetry_frames: tuple[RenderTelemetryFrameSnapshot, ...] = ()


def run_replay_render_benchmark(
    replay: Replay,
    *,
    replay_path: Path,
    base_dir: Path,
    assets_dir: Path | None = None,
    width: int | None = None,
    height: int | None = None,
    runs: int = 5,
    warmup_runs: int = 1,
    max_ticks: int | None = None,
    trace_rng: bool = False,
    profile: bool = False,
    profile_sort: ProfileSortKey = "cumtime",
    top: int = 20,
    profile_out: Path | None = None,
    mute_audio: bool = True,
    render_telemetry: bool = False,
    render_telemetry_out: Path | None = None,
    render_charts_out_dir: Path | None = None,
    rtx: bool = False,
    show_progress: bool = False,
) -> ReplayBenchmarkResult:
    from grim.config import ensure_crimson_cfg
    from grim.console import create_console
    from grim.raylib_api import rl

    from ...assets_fetch import download_missing_paqs

    _validate_args(runs=runs, warmup_runs=warmup_runs, top=top)
    telemetry_requested = bool(
        render_telemetry
        or render_telemetry_out is not None
        or render_charts_out_dir is not None,
    )

    baseline_result = build_verify_playback_driver(
        replay,
        max_ticks=max_ticks,
        trace_rng=bool(trace_rng),
    ).run()

    runtime_base_dir = Path(base_dir)
    runtime_assets_dir = Path(assets_dir) if assets_dir is not None else runtime_base_dir
    runtime_base_dir.mkdir(parents=True, exist_ok=True)
    cfg = ensure_crimson_cfg(runtime_base_dir)
    if bool(mute_audio):
        cfg.set_bool_value("sound_disable", True)
        cfg.set_bool_value("music_disable", True)

    render_width = int(width) if width is not None else cfg.screen_width
    render_height = int(height) if height is not None else cfg.screen_height
    if int(render_width) <= 0 or int(render_height) <= 0:
        raise ReplayBenchmarkError(
            f"invalid render resolution: {render_width}x{render_height}; width/height must be > 0",
        )

    console = create_console(runtime_base_dir, assets_dir=runtime_assets_dir)
    download_missing_paqs(runtime_assets_dir, console)
    ctx = ViewContext(assets_dir=runtime_assets_dir, preserve_bugs=False)

    config_flags = rl.ConfigFlags.FLAG_WINDOW_HIDDEN
    if int(config_flags) != 0:
        rl.set_config_flags(int(config_flags))
    try:
        rl.init_window(int(render_width), int(render_height), f"Replay Benchmark - {Path(replay_path).name}")
    except RuntimeError as exc:
        raise ReplayBenchmarkError(f"render benchmark could not initialize window: {exc}") from exc

    tick_total = len(replay.ticks)
    if max_ticks is not None:
        tick_total = min(tick_total, max(0, int(max_ticks)))

    progress: _ProgressBarLike | None = None
    try:
        planned_steps = int(warmup_runs) + int(runs) + (1 if bool(profile) else 0) + (1 if telemetry_requested else 0)
        if bool(show_progress) and planned_steps > 0:
            progress = cast(
                _ProgressBarLike,
                tqdm(
                    total=planned_steps,
                    unit="run",
                    desc="render benchmark",
                    leave=False,
                ),
            )

        def _run_once_with_tick_progress(
            *,
            tick_desc: str,
            rtx: bool,
            telemetry_session: RenderTelemetrySession | None = None,
        ) -> _RenderOnceResult:
            tick_progress: _ProgressBarLike | None = None
            completed_ticks = 0
            tick_callback: Callable[[int], None] | None = None
            completed_run = False
            if bool(show_progress) and tick_total > 0:
                tick_progress_bar = cast(
                    _ProgressBarLike,
                    tqdm(
                        total=tick_total,
                        unit="tick",
                        desc=str(tick_desc),
                        leave=False,
                    ),
                )
                tick_progress = tick_progress_bar

                def _on_tick(tick_index: int) -> None:
                    nonlocal completed_ticks
                    target_tick = max(0, min(tick_total, int(tick_index)))
                    if target_tick <= completed_ticks:
                        return
                    tick_progress_bar.update(target_tick - completed_ticks)
                    completed_ticks = target_tick

                tick_callback = _on_tick

            try:
                result = _run_render_once(
                    ctx=ctx,
                    replay_path=replay_path,
                    cfg=cfg,
                    console=console,
                    max_ticks=max_ticks,
                    trace_rng=bool(trace_rng),
                    rtx=bool(rtx),
                    telemetry_session=telemetry_session,
                    tick_progress_callback=tick_callback,
                )
                completed_run = True
                return result
            finally:
                if tick_progress is not None:
                    if completed_run and completed_ticks < tick_total:
                        tick_progress.update(tick_total - completed_ticks)
                    tick_progress.close()

        for _ in range(int(warmup_runs)):
            _run_once_with_tick_progress(
                tick_desc="render ticks warmup",
                rtx=bool(rtx),
            )
            if progress is not None:
                progress.update(1)
                progress.set_postfix_str("phase=warmup", refresh=False)

        samples: list[BenchmarkSample] = []
        for sample_idx in range(int(runs)):
            start_ns = time.perf_counter_ns()
            measured = _run_once_with_tick_progress(
                tick_desc=f"render ticks sample {sample_idx + 1}/{int(runs)}",
                rtx=bool(rtx),
            )
            elapsed_ns = max(1, int(time.perf_counter_ns()) - int(start_ns))
            wall_ms = float(elapsed_ns) / 1_000_000.0
            wall_s = float(elapsed_ns) / 1_000_000_000.0
            ticks_per_second = float(measured.run_result.ticks) / wall_s
            realtime_x = float(measured.run_result.elapsed_ms) / wall_ms
            samples.append(
                BenchmarkSample(
                    wall_ms=float(wall_ms),
                    ticks_per_second=float(ticks_per_second),
                    realtime_x=float(realtime_x),
                ),
            )
            _assert_consistent_run_result(
                baseline_result,
                measured.run_result,
                where=f"render run {sample_idx + 1}",
            )
            if progress is not None:
                progress.update(1)
                progress.set_postfix_str(f"phase=measure sample={sample_idx + 1}/{int(runs)}", refresh=False)

        profile_result: ReplayProfileResult | None = None
        if bool(profile):
            prof = cProfile.Profile()
            prof.enable()
            profiled = _run_once_with_tick_progress(
                tick_desc="render ticks profile",
                rtx=bool(rtx),
            )
            prof.disable()
            _assert_consistent_run_result(
                baseline_result,
                profiled.run_result,
                where="render profiled run",
            )

            if profile_out is not None:
                out_path = Path(profile_out)
                out_path.parent.mkdir(parents=True, exist_ok=True)
                prof.dump_stats(str(out_path))

            source, hotspots = _extract_hotspots(prof, sort_key=profile_sort, top=int(top))
            profile_result = ReplayProfileResult(
                sort=profile_sort,
                top=int(top),
                source=source,
                hotspots=tuple(hotspots),
            )
            if progress is not None:
                progress.update(1)
                progress.set_postfix_str("phase=profile", refresh=False)

        telemetry_result: ReplayRenderTelemetryResult | None = None
        if telemetry_requested:
            telemetry_session = RenderTelemetrySession()
            with telemetry_session:
                collected = _run_once_with_tick_progress(
                    tick_desc="render ticks telemetry",
                    rtx=bool(rtx),
                    telemetry_session=telemetry_session,
                )
            _assert_consistent_run_result(
                baseline_result,
                collected.run_result,
                where="render telemetry run",
            )

            frames = tuple(_convert_telemetry_frame(frame) for frame in collected.telemetry_frames)
            summary = _summarize_render_telemetry(frames=frames)

            telemetry_json_path: Path | None = None
            if render_telemetry_out is not None:
                telemetry_json_path = Path(render_telemetry_out)
                telemetry_json_path.parent.mkdir(parents=True, exist_ok=True)
                payload = {
                    "frames": [msgspec.to_builtins(frame) for frame in frames],
                    "summary": msgspec.to_builtins(summary),
                }
                telemetry_json_path.write_text(
                    json.dumps(payload, indent=2, sort_keys=True) + "\n",
                    encoding="utf-8",
                )

            chart_paths: dict[str, Path] = {}
            if render_charts_out_dir is not None:
                chart_paths = write_render_telemetry_charts(
                    frames=list(frames),
                    out_dir=Path(render_charts_out_dir),
                    telemetry_json_path=telemetry_json_path,
                )

            artifacts = ReplayRenderTelemetryArtifacts(
                telemetry_json_path=(str(telemetry_json_path) if telemetry_json_path is not None else None),
                charts_dir=(str(Path(render_charts_out_dir)) if render_charts_out_dir is not None else None),
                frame_timing_svg=(str(chart_paths.get("frame_timing_svg")) if chart_paths.get("frame_timing_svg") else None),
                draw_calls_svg=(str(chart_paths.get("draw_calls_svg")) if chart_paths.get("draw_calls_svg") else None),
                pass_timing_stacked_svg=(
                    str(chart_paths.get("pass_timing_stacked_svg"))
                    if chart_paths.get("pass_timing_stacked_svg")
                    else None
                ),
                report_md=(str(chart_paths.get("report_md")) if chart_paths.get("report_md") else None),
            )
            telemetry_result = ReplayRenderTelemetryResult(
                frames=frames,
                summary=summary,
                artifacts=artifacts,
                preview=tuple(frames[:10]),
            )
            if progress is not None:
                progress.update(1)
                progress.set_postfix_str("phase=telemetry", refresh=False)
    finally:
        if progress is not None:
            progress.close()
        rl.close_window()

    wall_values = [sample.wall_ms for sample in samples]
    tps_values = [sample.ticks_per_second for sample in samples]
    realtime_values = [sample.realtime_x for sample in samples]

    return ReplayBenchmarkResult(
        run_result=baseline_result,
        samples=tuple(samples),
        wall_ms=_aggregate(wall_values),
        ticks_per_second=_aggregate(tps_values),
        realtime_x=_aggregate(realtime_values),
        profile=profile_result,
        render_telemetry=telemetry_result,
    )


def run_replay_benchmark(
    replay: Replay,
    *,
    runs: int = 5,
    warmup_runs: int = 1,
    max_ticks: int | None = None,
    trace_rng: bool = False,
    profile: bool = False,
    profile_sort: ProfileSortKey = "cumtime",
    top: int = 20,
    profile_out: Path | None = None,
    show_progress: bool = False,
) -> ReplayBenchmarkResult:
    _validate_args(runs=runs, warmup_runs=warmup_runs, top=top)

    tick_total = len(replay.ticks)
    if max_ticks is not None:
        tick_total = min(tick_total, max(0, int(max_ticks)))

    progress: _ProgressBarLike | None = None
    try:
        planned_steps = int(warmup_runs) + int(runs) + (1 if bool(profile) else 0)
        if bool(show_progress) and planned_steps > 0:
            progress = cast(
                _ProgressBarLike,
                tqdm(
                    total=planned_steps,
                    unit="run",
                    desc="headless benchmark",
                    leave=False,
                ),
            )

        def _run_once_with_tick_progress(*, tick_desc: str) -> RunResult:
            tick_progress: _ProgressBarLike | None = None
            completed_ticks = 0
            tick_callback: Callable[[int], None] | None = None
            completed_run = False
            if bool(show_progress) and tick_total > 0:
                tick_progress_bar = cast(
                    _ProgressBarLike,
                    tqdm(
                        total=tick_total,
                        unit="tick",
                        desc=str(tick_desc),
                        leave=False,
                    ),
                )
                tick_progress = tick_progress_bar

                def _on_tick(tick_index: int) -> None:
                    nonlocal completed_ticks
                    target_tick = max(0, min(tick_total, int(tick_index)))
                    if target_tick <= completed_ticks:
                        return
                    tick_progress_bar.update(target_tick - completed_ticks)
                    completed_ticks = target_tick

                tick_callback = _on_tick

            try:
                driver = build_verify_playback_driver(
                    replay,
                    max_ticks=max_ticks,
                    trace_rng=bool(trace_rng),
                )
                result = driver.run(
                    hooks=PlaybackWalkHooks(on_progress=tick_callback),
                )
                completed_run = True
                return result
            finally:
                if tick_progress is not None:
                    if completed_run and completed_ticks < tick_total:
                        tick_progress.update(tick_total - completed_ticks)
                    tick_progress.close()

        for _ in range(int(warmup_runs)):
            _run_once_with_tick_progress(tick_desc="headless ticks warmup")
            if progress is not None:
                progress.update(1)
                progress.set_postfix_str("phase=warmup", refresh=False)

        baseline_result: RunResult | None = None
        samples: list[BenchmarkSample] = []
        for sample_idx in range(int(runs)):
            start_ns = time.perf_counter_ns()
            result = _run_once_with_tick_progress(
                tick_desc=f"headless ticks sample {sample_idx + 1}/{int(runs)}",
            )
            elapsed_ns = max(1, int(time.perf_counter_ns()) - int(start_ns))
            wall_ms = float(elapsed_ns) / 1_000_000.0
            wall_s = float(elapsed_ns) / 1_000_000_000.0
            ticks_per_second = float(result.ticks) / wall_s
            realtime_x = float(result.elapsed_ms) / wall_ms
            samples.append(
                BenchmarkSample(
                    wall_ms=float(wall_ms),
                    ticks_per_second=float(ticks_per_second),
                    realtime_x=float(realtime_x),
                ),
            )
            if baseline_result is None:
                baseline_result = result
            else:
                _assert_consistent_run_result(baseline_result, result, where=f"measured run {sample_idx + 1}")
            if progress is not None:
                progress.update(1)
                progress.set_postfix_str(f"phase=measure sample={sample_idx + 1}/{int(runs)}", refresh=False)

        assert baseline_result is not None
        profile_result: ReplayProfileResult | None = None
        if bool(profile):
            prof = cProfile.Profile()
            prof.enable()
            prof_result = _run_once_with_tick_progress(tick_desc="headless ticks profile")
            prof.disable()
            _assert_consistent_run_result(baseline_result, prof_result, where="profiled run")

            if profile_out is not None:
                out_path = Path(profile_out)
                out_path.parent.mkdir(parents=True, exist_ok=True)
                prof.dump_stats(str(out_path))

            source, hotspots = _extract_hotspots(prof, sort_key=profile_sort, top=int(top))
            profile_result = ReplayProfileResult(
                sort=profile_sort,
                top=int(top),
                source=source,
                hotspots=tuple(hotspots),
            )
            if progress is not None:
                progress.update(1)
                progress.set_postfix_str("phase=profile", refresh=False)
    finally:
        if progress is not None:
            progress.close()

    wall_values = [sample.wall_ms for sample in samples]
    tps_values = [sample.ticks_per_second for sample in samples]
    realtime_values = [sample.realtime_x for sample in samples]

    return ReplayBenchmarkResult(
        run_result=baseline_result,
        samples=tuple(samples),
        wall_ms=_aggregate(wall_values),
        ticks_per_second=_aggregate(tps_values),
        realtime_x=_aggregate(realtime_values),
        profile=profile_result,
    )


def _validate_args(*, runs: int, warmup_runs: int, top: int) -> None:
    if int(runs) < 1:
        raise ReplayBenchmarkError("runs must be >= 1")
    if int(warmup_runs) < 0:
        raise ReplayBenchmarkError("warmup_runs must be >= 0")
    if int(top) < 1:
        raise ReplayBenchmarkError("top must be >= 1")


def _run_render_once(
    *,
    ctx: ViewContext,
    replay_path: Path,
    cfg: CrimsonConfig,
    console: ConsoleState,
    max_ticks: int | None,
    trace_rng: bool,
    rtx: bool,
    telemetry_session: RenderTelemetrySession | None = None,
    tick_progress_callback: Callable[[int], None] | None = None,
) -> _RenderOnceResult:
    mode = ReplayPlaybackMode(
        ctx,
        replay_path=Path(replay_path),
        config=cfg,
        console=console,
        max_ticks=max_ticks,
        trace_rng=bool(trace_rng),
        rtx=bool(rtx),
    )
    mode.open()
    try:
        replay = mode._replay
        if replay is None:
            raise ReplayBenchmarkError("render benchmark failed: replay playback did not initialize replay state")

        step_dt = float(mode._dt)
        if step_dt <= 0.0:
            step_dt = 1.0 / 60.0

        frame_index = 0
        while not bool(mode.finished):
            tick_before = int(mode.tick_index)
            if telemetry_session is not None:
                telemetry_session.begin_frame(
                    frame_index=int(frame_index),
                    tick_index_before_update=int(tick_before),
                )

            frame_start_ns = time.perf_counter_ns()
            update_start_ns = time.perf_counter_ns()
            mode.update(float(step_dt))
            update_ns = max(0, int(time.perf_counter_ns()) - int(update_start_ns))

            draw_start_ns = time.perf_counter_ns()
            rl.begin_drawing()
            try:
                mode.draw()
            finally:
                rl.end_drawing()
            draw_ns = max(0, int(time.perf_counter_ns()) - int(draw_start_ns))
            frame_ns = max(0, int(time.perf_counter_ns()) - int(frame_start_ns))

            tick_after = int(mode.tick_index)
            if tick_progress_callback is not None:
                tick_progress_callback(int(tick_after))
            if telemetry_session is not None:
                telemetry_session.end_frame(
                    tick_index_after_update=int(tick_after),
                    update_ms=float(update_ns) / 1_000_000.0,
                    draw_ms=float(draw_ns) / 1_000_000.0,
                    frame_ms=float(frame_ns) / 1_000_000.0,
                )

            if bool(mode.close_requested):
                raise ReplayBenchmarkError("render benchmark aborted: replay playback requested close")

            frame_index += 1

        return _RenderOnceResult(
            run_result=_run_result_for_replay_mode(mode=mode),
            telemetry_frames=(telemetry_session.frames if telemetry_session is not None else ()),
        )
    finally:
        mode.close()


def _run_result_for_replay_mode(*, mode: ReplayPlaybackMode) -> RunResult:
    driver = mode._driver
    if driver is None:
        raise ReplayBenchmarkError("render benchmark failed: replay driver was not available")
    return driver.build_run_result(ticks=int(mode.tick_index))


def _assert_consistent_run_result(expected: RunResult, actual: RunResult, *, where: str) -> None:
    expected_values = (
        (expected.game_mode_id, actual.game_mode_id, "game_mode_id"),
        (expected.tick_rate, actual.tick_rate, "tick_rate"),
        (expected.ticks, actual.ticks, "ticks"),
        (expected.elapsed_ms, actual.elapsed_ms, "elapsed_ms"),
        (expected.score_xp, actual.score_xp, "score_xp"),
        (expected.creature_kill_count, actual.creature_kill_count, "creature_kill_count"),
        (expected.most_used_weapon_id, actual.most_used_weapon_id, "most_used_weapon_id"),
        (expected.shots_fired, actual.shots_fired, "shots_fired"),
        (expected.shots_hit, actual.shots_hit, "shots_hit"),
        (expected.rng_state, actual.rng_state, "rng_state"),
    )
    for exp, got, field_name in expected_values:
        if int(exp) != int(got):
            raise ReplayBenchmarkError(
                "non-deterministic replay result across runs: "
                f"{field_name} expected={exp} actual={got} ({where})",
            )


def _aggregate(values: list[float]) -> BenchmarkAggregate:
    if not values:
        raise ReplayBenchmarkError("benchmark produced no samples")
    sorted_values = sorted(float(value) for value in values)
    mean_value = float(statistics.fmean(sorted_values))
    stdev_value = float(statistics.stdev(sorted_values)) if len(sorted_values) >= 2 else 0.0
    return BenchmarkAggregate(
        min=float(sorted_values[0]),
        p50=_percentile(sorted_values, 0.50),
        mean=float(mean_value),
        p95=_percentile(sorted_values, 0.95),
        max=float(sorted_values[-1]),
        stdev=float(stdev_value),
    )


def _aggregate_or_zero(values: list[float]) -> BenchmarkAggregate:
    if not values:
        return BenchmarkAggregate(min=0.0, p50=0.0, mean=0.0, p95=0.0, max=0.0, stdev=0.0)
    return _aggregate(values)


def _percentile(sorted_values: list[float], ratio: float) -> float:
    if not sorted_values:
        raise ReplayBenchmarkError("cannot compute percentile of empty values")
    if len(sorted_values) == 1:
        return float(sorted_values[0])
    clamped = min(1.0, max(0.0, float(ratio)))
    pos = (len(sorted_values) - 1) * clamped
    lo = int(math.floor(pos))
    hi = int(math.ceil(pos))
    if lo == hi:
        return float(sorted_values[lo])
    frac = float(pos) - float(lo)
    return float(sorted_values[lo] * (1.0 - frac) + sorted_values[hi] * frac)


def _extract_hotspots(
    profile: cProfile.Profile,
    *,
    sort_key: ProfileSortKey,
    top: int,
) -> tuple[HotspotSource, list[ReplayProfileHotspot]]:
    stats = cast(_ProfileStatsLike, pstats.Stats(profile))
    rows: list[ReplayProfileHotspot] = []
    for key, values in stats.stats.items():
        file_name, line_number, function_name = key
        primitive_calls, total_calls, tottime, cumtime, _callers = values
        rows.append(
            ReplayProfileHotspot(
                file=str(file_name),
                line=int(line_number),
                function=str(function_name),
                primitive_calls=int(primitive_calls),
                total_calls=int(total_calls),
                tottime=float(tottime),
                cumtime=float(cumtime),
            ),
        )

    if str(sort_key) == "tottime":
        rows.sort(key=lambda row: float(row.tottime), reverse=True)
    else:
        rows.sort(key=lambda row: float(row.cumtime), reverse=True)

    project_rows = [row for row in rows if _is_project_hotspot_path(str(row.file))]
    if project_rows:
        return "project", project_rows[: int(top)]
    return "all", rows[: int(top)]


def _is_project_hotspot_path(path: str) -> bool:
    text = str(path).replace("\\", "/").lower()
    return (
        text.startswith("src/crimson/")
        or text.startswith("src/grim/")
        or text.startswith("crimson/")
        or text.startswith("grim/")
        or "/src/crimson/" in text
        or "/src/grim/" in text
    )


def _convert_telemetry_frame(frame: RenderTelemetryFrameSnapshot) -> ReplayRenderTelemetryFrame:
    return ReplayRenderTelemetryFrame(
        frame_index=int(frame.frame_index),
        tick_index_before_update=int(frame.tick_index_before_update),
        tick_index_after_update=int(frame.tick_index_after_update),
        update_ms=float(frame.update_ms),
        draw_ms=float(frame.draw_ms),
        frame_ms=float(frame.frame_ms),
        draw_calls_total=int(frame.draw_calls_total),
        draw_calls_by_api={str(key): int(value) for key, value in frame.draw_calls_by_api.items()},
        draw_calls_by_pass={str(key): int(value) for key, value in frame.draw_calls_by_pass.items()},
        pass_ms={str(key): float(value) for key, value in frame.pass_ms.items()},
    )


def _top_ticks(
    *,
    frames: tuple[ReplayRenderTelemetryFrame, ...],
    top_n: int,
    key_fn: Callable[[ReplayRenderTelemetryFrame], float],
) -> tuple[ReplayRenderTelemetryTopTick, ...]:
    ordered = sorted(frames, key=key_fn, reverse=True)
    selected = ordered[: max(0, int(top_n))]
    return tuple(
        ReplayRenderTelemetryTopTick(
            tick_index=int(frame.tick_index_after_update),
            frame_index=int(frame.frame_index),
            value=float(key_fn(frame)),
        )
        for frame in selected
    )


def _summarize_render_telemetry(*, frames: tuple[ReplayRenderTelemetryFrame, ...]) -> ReplayRenderTelemetrySummary:
    frame_ms_values = [float(frame.frame_ms) for frame in frames]
    update_ms_values = [float(frame.update_ms) for frame in frames]
    draw_ms_values = [float(frame.draw_ms) for frame in frames]
    draw_calls_values = [float(frame.draw_calls_total) for frame in frames]

    return ReplayRenderTelemetrySummary(
        frame_ms=_aggregate_or_zero(frame_ms_values),
        update_ms=_aggregate_or_zero(update_ms_values),
        draw_ms=_aggregate_or_zero(draw_ms_values),
        draw_calls_total=_aggregate_or_zero(draw_calls_values),
        top_draw_ms_ticks=_top_ticks(frames=frames, top_n=5, key_fn=lambda row: float(row.draw_ms)),
        top_frame_ms_ticks=_top_ticks(frames=frames, top_n=5, key_fn=lambda row: float(row.frame_ms)),
        top_draw_calls_ticks=_top_ticks(frames=frames, top_n=5, key_fn=lambda row: float(row.draw_calls_total)),
    )



from __future__ import annotations

import hashlib
import io
import struct
from collections.abc import Iterable, Iterator, Sequence
from pathlib import Path

import msgspec
import zstandard as zstd

from .schema import (
    CHUNK_FLAG_MSGPACK,
    CHUNK_FLAG_ZSTD,
    CHUNK_KIND_FOOTER,
    CHUNK_KIND_META,
    CHUNK_KIND_TICK,
    CHUNK_KINDS,
    DEFAULT_CHUNK_FLAGS,
    SUPPORTED_TRACE_SCHEMA_VERSIONS,
    TRACE_FORMAT_VERSION,
    TRACE_MAGIC,
    TRAILER_MAGIC,
    TickBlock,
    TickBlockIndexEntry,
    TickRecord,
    TraceFooter,
    TraceMeta,
)

_FILE_HEADER_STRUCT = struct.Struct("<I")
_CHUNK_HEADER_STRUCT = struct.Struct("<4siiIIIQ")
_TRAILER_STRUCT = struct.Struct("<8sQ")
_META_MAGIC_LEN = len(TRACE_MAGIC)

_ENCODER = msgspec.msgpack.Encoder()
_META_DECODER = msgspec.msgpack.Decoder(type=TraceMeta)
_BLOCK_DECODER = msgspec.msgpack.Decoder(type=TickBlock)
_FOOTER_DECODER = msgspec.msgpack.Decoder(type=TraceFooter)


class TraceError(ValueError):
    pass


class TraceSummary(msgspec.Struct, frozen=True):
    meta: TraceMeta
    footer: TraceFooter


def _checksum64(data: bytes) -> int:
    digest = hashlib.blake2b(data, digest_size=8).digest()
    return int.from_bytes(digest, byteorder="little", signed=False)


def _compress(payload: bytes) -> bytes:
    compressor = zstd.ZstdCompressor(level=3)
    return compressor.compress(payload)


def _decompress(payload: bytes) -> bytes:
    decompressor = zstd.ZstdDecompressor()
    return decompressor.decompress(payload)


def _chunk_kind_bytes(kind: str) -> bytes:
    kind_bytes = kind.encode("ascii")
    if kind_bytes not in CHUNK_KINDS:
        raise TraceError(f"unsupported chunk kind: {kind!r}")
    return kind_bytes


def _write_chunk(
    stream: io.BufferedWriter,
    *,
    kind: str,
    start_tick: int,
    end_tick: int,
    payload: bytes,
) -> TickBlockIndexEntry:
    kind_bytes = _chunk_kind_bytes(kind)
    compressed = _compress(payload)
    checksum = _checksum64(payload)
    offset = stream.tell()
    header = _CHUNK_HEADER_STRUCT.pack(
        kind_bytes,
        start_tick,
        end_tick,
        DEFAULT_CHUNK_FLAGS,
        len(compressed),
        len(payload),
        checksum,
    )
    stream.write(header)
    stream.write(compressed)
    return TickBlockIndexEntry(
        start_tick=start_tick,
        end_tick=end_tick,
        file_offset=offset,
        compressed_len=len(compressed),
        uncompressed_len=len(payload),
        checksum=checksum,
    )


def _write_trace_from_iter(
    path: Path,
    *,
    meta: TraceMeta,
    ticks: Iterable[TickRecord],
    chunk_ticks: int,
) -> TraceSummary:
    path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)
    chunk_size = max(1, int(chunk_ticks))

    with path.open("wb") as handle:
        handle.write(TRACE_MAGIC)
        handle.write(_FILE_HEADER_STRUCT.pack(TRACE_FORMAT_VERSION))

        _write_chunk(
            handle,
            kind=CHUNK_KIND_META,
            start_tick=-1,
            end_tick=-1,
            payload=_ENCODER.encode(meta),
        )

        tick_indices: list[TickBlockIndexEntry] = []
        channel_counts: dict[str, int] = {}
        first_tick: int | None = None
        last_tick: int | None = None
        tick_count = 0
        current_block: list[TickRecord] = []

        def flush_block() -> None:
            nonlocal first_tick, last_tick, tick_count
            if not current_block:
                return
            block = TickBlock(
                start_tick=int(current_block[0].tick_index),
                end_tick=int(current_block[-1].tick_index),
                ticks=list(current_block),
            )
            tick_indices.append(
                _write_chunk(
                    handle,
                    kind=CHUNK_KIND_TICK,
                    start_tick=int(block.start_tick),
                    end_tick=int(block.end_tick),
                    payload=_ENCODER.encode(block),
                ),
            )
            for row in current_block:
                row_tick = int(row.tick_index)
                if first_tick is None:
                    first_tick = row_tick
                if last_tick is not None and row_tick < int(last_tick):
                    raise TraceError(
                        f"tick rows must be in non-decreasing order, got {row_tick} after {int(last_tick)}",
                    )
                last_tick = row_tick
                tick_count += 1
                for channel_name in row.channels:
                    channel_counts[channel_name] = channel_counts.get(channel_name, 0) + 1
            current_block.clear()

        for tick in ticks:
            current_block.append(tick)
            if len(current_block) >= int(chunk_size):
                flush_block()
        flush_block()

        if int(tick_count) <= 0:
            raise TraceError("trace must contain at least one tick")
        if first_tick is None or last_tick is None:
            raise TraceError("trace footer tick bounds are missing")
        footer = TraceFooter(
            trace_format_version=TRACE_FORMAT_VERSION,
            tick_blocks=tick_indices,
            tick_count=int(tick_count),
            first_tick=int(first_tick),
            last_tick=int(last_tick),
            channel_counts={key: value for key, value in sorted(channel_counts.items())},
        )
        footer_payload = _ENCODER.encode(footer)
        footer_index = _write_chunk(
            handle,
            kind=CHUNK_KIND_FOOTER,
            start_tick=-1,
            end_tick=-1,
            payload=footer_payload,
        )
        handle.write(
            _TRAILER_STRUCT.pack(
                TRAILER_MAGIC,
                footer_index.file_offset,
            ),
        )

    return TraceSummary(meta=meta, footer=footer)


def write_trace_iter(
    path: Path,
    *,
    meta: TraceMeta,
    ticks: Iterable[TickRecord],
    chunk_ticks: int = 256,
) -> TraceSummary:
    return _write_trace_from_iter(
        path,
        meta=meta,
        ticks=ticks,
        chunk_ticks=max(1, int(chunk_ticks)),
    )


def _chunk_payload_from_file(stream: io.BufferedReader, *, offset: int) -> tuple[str, int, int, bytes]:
    stream.seek(offset)
    header = stream.read(_CHUNK_HEADER_STRUCT.size)
    if len(header) != _CHUNK_HEADER_STRUCT.size:
        raise TraceError("truncated chunk header")
    kind_bytes, start_tick, end_tick, flags, compressed_len, raw_len, checksum = _CHUNK_HEADER_STRUCT.unpack(header)
    if kind_bytes not in CHUNK_KINDS:
        raise TraceError(f"unknown chunk kind: {kind_bytes!r}")
    payload_compressed = stream.read(compressed_len)
    if len(payload_compressed) != compressed_len:
        raise TraceError("truncated chunk payload")
    payload = payload_compressed
    if flags & CHUNK_FLAG_ZSTD:
        payload = _decompress(payload_compressed)
    if flags & CHUNK_FLAG_MSGPACK == 0:
        raise TraceError("unsupported chunk payload encoding")
    if len(payload) != raw_len:
        raise TraceError("chunk size mismatch")
    if _checksum64(payload) != checksum:
        raise TraceError("chunk checksum mismatch")
    return kind_bytes.decode("ascii"), start_tick, end_tick, payload


def write_trace(
    path: Path,
    *,
    meta: TraceMeta,
    ticks: Sequence[TickRecord],
    chunk_ticks: int = 256,
) -> TraceSummary:
    sorted_ticks = sorted((tick for tick in ticks), key=lambda row: int(row.tick_index))
    return _write_trace_from_iter(
        path,
        meta=meta,
        ticks=sorted_ticks,
        chunk_ticks=max(1, int(chunk_ticks)),
    )


def _load_meta_at_offset(stream: io.BufferedReader, *, offset: int) -> TraceMeta:
    kind, _start, _end, payload = _chunk_payload_from_file(stream, offset=offset)
    if kind != CHUNK_KIND_META:
        raise TraceError("invalid trace meta chunk")
    meta = _META_DECODER.decode(payload)
    if int(meta.trace_schema_version) not in SUPPORTED_TRACE_SCHEMA_VERSIONS:
        supported = ", ".join(str(version) for version in sorted(SUPPORTED_TRACE_SCHEMA_VERSIONS))
        raise TraceError(
            f"unsupported trace schema version: {meta.trace_schema_version} (supported: {supported})",
        )
    return meta


def _load_footer_and_meta_offsets(path: Path) -> tuple[int, int]:
    path = Path(path)
    with path.open("rb") as handle:
        file_header = handle.read(_META_MAGIC_LEN + _FILE_HEADER_STRUCT.size)
        if len(file_header) != _META_MAGIC_LEN + _FILE_HEADER_STRUCT.size:
            raise TraceError(f"invalid trace file: {path}")
        magic = file_header[:_META_MAGIC_LEN]
        if magic != TRACE_MAGIC:
            raise TraceError(f"unsupported trace magic: {path}")
        (format_version,) = _FILE_HEADER_STRUCT.unpack(file_header[_META_MAGIC_LEN:])
        if format_version != TRACE_FORMAT_VERSION:
            raise TraceError(f"unsupported trace format version: {format_version}")

        handle.seek(0, io.SEEK_END)
        end_pos = handle.tell()
        if end_pos < (_META_MAGIC_LEN + _FILE_HEADER_STRUCT.size + _TRAILER_STRUCT.size):
            raise TraceError("invalid trace size")
        handle.seek(end_pos - _TRAILER_STRUCT.size)
        trailer_raw = handle.read(_TRAILER_STRUCT.size)
        trailer_magic, footer_offset = _TRAILER_STRUCT.unpack(trailer_raw)
        if trailer_magic != TRAILER_MAGIC:
            raise TraceError("missing trace trailer")
        return footer_offset, _META_MAGIC_LEN + _FILE_HEADER_STRUCT.size


def load_trace_meta(path: Path) -> TraceMeta:
    path = Path(path)
    footer_offset, meta_offset = _load_footer_and_meta_offsets(path)
    _ = footer_offset
    with path.open("rb") as handle:
        return _load_meta_at_offset(handle, offset=meta_offset)


class TraceReader:
    def __init__(self, path: Path) -> None:
        self.path = Path(path)
        footer_offset, meta_offset = _load_footer_and_meta_offsets(self.path)
        self._handle = self.path.open("rb")
        self.meta = _load_meta_at_offset(self._handle, offset=meta_offset)
        footer_kind, _start, _end, footer_payload = _chunk_payload_from_file(self._handle, offset=footer_offset)
        if footer_kind != CHUNK_KIND_FOOTER:
            raise TraceError("invalid trace footer chunk")
        self.footer = _FOOTER_DECODER.decode(footer_payload)
        if int(self.footer.tick_count) <= 0:
            raise TraceError("invalid trace footer tick_count")
        if int(self.footer.first_tick) < 0 or int(self.footer.last_tick) < 0:
            raise TraceError("invalid trace footer tick bounds")
        if int(self.footer.first_tick) > int(self.footer.last_tick):
            raise TraceError("invalid trace footer tick order")
        if len(self.footer.tick_blocks) <= 0:
            raise TraceError("invalid trace footer block index")
        self._block_cache: dict[int, TickBlock] = {}

    def close(self) -> None:
        if not self._handle.closed:
            self._handle.close()

    def __enter__(self) -> "TraceReader":
        return self

    def __exit__(self, *_args: object) -> None:
        self.close()

    def _load_block(self, entry: TickBlockIndexEntry) -> TickBlock:
        cache_key = entry.file_offset
        cached = self._block_cache.get(cache_key)
        if cached is not None:
            return cached
        kind, _start, _end, payload = _chunk_payload_from_file(self._handle, offset=cache_key)
        if kind != CHUNK_KIND_TICK:
            raise TraceError("trace index points at non-tick chunk")
        block = _BLOCK_DECODER.decode(payload)
        self._block_cache[cache_key] = block
        return block

    def tick(self, tick_index: int) -> TickRecord | None:
        tick = tick_index
        for entry in self.footer.tick_blocks:
            if entry.start_tick <= tick <= entry.end_tick:
                block = self._load_block(entry)
                for row in block.ticks:
                    if row.tick_index == tick:
                        return row
                return None
        return None

    def iter_ticks(self, *, tick_start: int | None = None, tick_end: int | None = None) -> Iterator[TickRecord]:
        start = tick_start
        end = tick_end
        for entry in self.footer.tick_blocks:
            if start is not None and entry.end_tick < start:
                continue
            if end is not None and entry.start_tick > end:
                continue
            block = self._load_block(entry)
            for row in block.ticks:
                tick = row.tick_index
                if start is not None and tick < start:
                    continue
                if end is not None and tick > end:
                    continue
                yield row

    def all_ticks(self) -> list[TickRecord]:
        return list(self.iter_ticks())


def load_trace(path: Path) -> tuple[TraceMeta, list[TickRecord], TraceFooter]:
    with TraceReader(Path(path)) as trace:
        return trace.meta, trace.all_ticks(), trace.footer


def iter_trace_ticks(path: Path, *, tick_start: int | None = None, tick_end: int | None = None) -> Iterable[TickRecord]:
    with TraceReader(Path(path)) as trace:
        yield from trace.iter_ticks(tick_start=tick_start, tick_end=tick_end)



from __future__ import annotations

import math
from collections.abc import Sequence

from ..sim.input import PlayerInput
from ..sim.input_providers import GameCommand
from .input_codec import pack_tick_inputs
from .types import (
    REPLAY_FORMAT_VERSION,
    Replay,
    ReplayHeader,
    ReplayTick,
)


class ReplayRecorder:
    def __init__(self, header: ReplayHeader) -> None:
        if header.replay_format_version != REPLAY_FORMAT_VERSION:
            raise ValueError(f"unsupported replay format version: {header.replay_format_version}")
        tick_rate = header.tick_rate
        if tick_rate <= 0:
            raise ValueError(f"invalid tick_rate: {tick_rate}")
        self._header = header
        self._default_dt = 1.0 / float(tick_rate)
        self._tick_index = 0
        self._ticks: list[ReplayTick] = []

    @property
    def header(self) -> ReplayHeader:
        return self._header

    @property
    def tick_index(self) -> int:
        return self._tick_index

    @property
    def recorded_tick_count(self) -> int:
        return len(self._ticks)

    def record_tick(
        self,
        inputs: Sequence[PlayerInput],
        *,
        commands: list[GameCommand] | None = None,
        dt: float | None = None,
    ) -> int:
        """Record a single simulation tick worth of inputs.

        Returns the tick index that was recorded.
        """

        if len(inputs) != self._header.player_count:
            raise ValueError(f"expected {self._header.player_count} player inputs, got {len(inputs)}")

        packed = pack_tick_inputs(inputs, quant=self._header.input_quantization)
        tick_dt = dt if dt is not None else self._default_dt
        if not math.isfinite(tick_dt) or tick_dt < 0.0:
            raise ValueError(f"dt must be finite and >= 0, got {tick_dt!r}")

        tick_index = self._tick_index
        self._ticks.append(ReplayTick(dt=tick_dt, inputs=packed, commands=list(commands) if commands else []))
        self._tick_index += 1
        return tick_index

    def finish(self) -> Replay:
        return Replay(header=self._header, ticks=self._ticks)



from __future__ import annotations

import math
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING

import msgspec

from grim.color import RGBA
from grim.geom import Vec2

from ..math_parity import NATIVE_TAU, f32, heading_from_delta_f32
from ..perks import PerkId
from ..perks.helpers import perk_active
from ..projectiles.runtime import SecondarySpawnSpec
from ..projectiles.types import ProjectileTemplateId, SecondaryProjectileTypeId
from ..sim.input import PlayerInput
from ..sim.state_types import GameplayState, PlayerState
from ..weapons import WEAPON_TABLE, WeaponId, weapon_entry_for_projectile_type_id
from .assign import player_start_reload, weapon_entry
from .fire_recipes import (
    MaskCenteredJitter,
    ModuloCenteredJitter,
    ModuloSpeedScale,
    MultiPlasmaFanMode,
    NoJitter,
    NoSpeedScale,
    ParticleStreamMode,
    PrimaryPelletsMode,
    SecondaryShotMode,
    SwarmerDumpMode,
    UseAimTargetHint,
    resolve_fire_recipe,
)
from .spawn import owner_ref_for_player, owner_ref_for_player_projectiles, travel_budget_for_type_id

if TYPE_CHECKING:
    from ..creatures.runtime import CreatureState

WEAPON_COUNT_SIZE = max(int(entry.weapon_id) for entry in WEAPON_TABLE) + 1

_NATIVE_FIRE_MUZZLE_SPRITES: dict[int, tuple[tuple[float, float, float], ...]] = {
    WeaponId.PISTOL: ((25.0, 1.0, 0.23), (15.0, 2.0, 0.213)),
    WeaponId.ASSAULT_RIFLE: ((25.0, 1.0, 0.23), (15.0, 2.0, 0.213)),
    WeaponId.SHOTGUN: ((25.0, 1.0, 0.25), (15.0, 2.0, 0.223)),
    WeaponId.SAWED_OFF_SHOTGUN: ((25.0, 1.0, 0.26), (15.0, 2.0, 0.233)),
    WeaponId.SUBMACHINE_GUN: ((25.0, 1.0, 0.23), (15.0, 2.0, 0.213)),
    WeaponId.GAUSS_GUN: ((25.0, 1.0, 0.33), (15.0, 2.0, 0.263)),
    WeaponId.ROCKET_LAUNCHER: ((25.0, 1.0, 0.34), (15.0, 2.0, 0.283)),
    WeaponId.SEEKER_ROCKETS: ((25.0, 1.0, 0.31), (15.0, 2.0, 0.243)),
    WeaponId.MINI_ROCKET_SWARMERS: ((25.0, 1.0, 0.34), (15.0, 2.0, 0.283)),
    WeaponId.ROCKET_MINIGUN: ((25.0, 1.0, 0.34),),
    WeaponId.JACKHAMMER: ((15.0, 2.0, 0.223),),
    WeaponId.SHRINKIFIER_5K: ((25.0, 1.0, 0.23), (15.0, 2.0, 0.213)),
    WeaponId.GAUSS_SHOTGUN: ((25.0, 1.0, 0.33), (15.0, 2.0, 0.263)),
}

_NATIVE_FIRE_MUZZLE_AFTER_PROJECTILE: frozenset[int] = frozenset(
    {
        WeaponId.PISTOL,
        WeaponId.SHRINKIFIER_5K,
    },
)


class WeaponFireCtx(msgspec.Struct):
    player: PlayerState
    input_state: PlayerInput
    dt: float
    state: GameplayState
    detail_preset: int = 5
    creatures: Sequence[CreatureState] | None = None
    players: Sequence[PlayerState] | None = None
    force_pre_swap_fire_gate: bool = False
    on_player_lethal: Callable[[PlayerState], None] | None = None


class WeaponFireResult(msgspec.Struct, frozen=True):
    fired: bool
    shot_count: int = 0
    ammo_cost: float = 0.0


def _spawn_native_fire_muzzle_sprites(
    *,
    state: GameplayState,
    weapon_id: int,
    muzzle: Vec2,
    aim_heading: float,
    fire_bullets_active: bool,
) -> None:
    if fire_bullets_active:
        specs: tuple[tuple[float, float, float], ...] = ((25.0, 1.0, 0.413),)
    else:
        specs = _NATIVE_FIRE_MUZZLE_SPRITES.get(int(weapon_id), ())
    if not specs:
        return

    for speed, scale, alpha in specs:
        state.sprite_effects.spawn(
            pos=muzzle,
            vel=Vec2.from_heading(aim_heading) * float(speed),
            scale=float(scale),
            color=RGBA(0.5, 0.5, 0.5, float(alpha)),
        )


def _native_shot_angle_with_jitter(
    *,
    aim: Vec2,
    player_pos: Vec2,
    spread_heat: float,
    rand: Callable[[], int],
) -> float:
    # `player_fire_weapon` computes jitter in float locals before `to_heading`.
    aim_dx = float(f32(float(aim.x) - float(player_pos.x)))
    aim_dy = float(f32(float(aim.y) - float(player_pos.y)))
    dist_sq = float(f32(float(f32(float(aim_dx) * float(aim_dx))) + float(f32(float(aim_dy) * float(aim_dy)))))
    dist = float(f32(math.sqrt(float(dist_sq))))
    max_offset = float(f32(float(f32(float(dist) * float(spread_heat))) * 0.5))

    dir_angle = float(f32(float(int(rand()) & 0x1FF) * (float(NATIVE_TAU) / 512.0)))
    mag = float(f32(float(int(rand()) & 0x1FF) * (1.0 / 512.0)))
    offset = float(f32(float(max_offset) * float(mag)))

    dir_x = float(f32(math.cos(float(dir_angle))))
    dir_y = float(f32(math.sin(float(dir_angle))))
    aim_jitter_x = float(f32(float(aim.x) + float(f32(float(dir_x) * float(offset)))))
    aim_jitter_y = float(f32(float(aim.y) + float(f32(float(dir_y) * float(offset)))))

    shot_dx = float(f32(float(aim_jitter_x) - float(player_pos.x)))
    shot_dy = float(f32(float(aim_jitter_y) - float(player_pos.y)))
    return float(heading_from_delta_f32(dx=float(shot_dx), dy=float(shot_dy)))


def _apply_pellet_jitter(
    *,
    shot_angle: float,
    rand: Callable[[], int],
    jitter_rule: NoJitter | ModuloCenteredJitter | MaskCenteredJitter,
) -> float:
    match jitter_rule:
        case NoJitter():
            return float(shot_angle)
        case ModuloCenteredJitter(modulo=modulo, center=center, step=step):
            return float(shot_angle) + float(int(rand()) % int(modulo) - int(center)) * float(step)
        case MaskCenteredJitter(mask=mask, center=center, step=step):
            return float(shot_angle) + float((int(rand()) & int(mask)) - int(center)) * float(step)


def _apply_speed_scale_rule(
    *,
    state: GameplayState,
    proj_id: int,
    speed_rule: NoSpeedScale | ModuloSpeedScale,
) -> None:
    match speed_rule:
        case NoSpeedScale():
            return
        case ModuloSpeedScale(base=base, modulo=modulo, step=step):
            state.projectiles.entries[int(proj_id)].speed_scale = float(base) + float(int(state.rng.rand()) % int(modulo)) * float(step)


def fire_weapon(ctx: WeaponFireCtx) -> WeaponFireResult:
    player = ctx.player
    input_state = ctx.input_state
    dt = float(ctx.dt)
    state = ctx.state
    detail_preset = int(ctx.detail_preset)
    creatures = ctx.creatures
    players = ctx.players
    force_pre_swap_fire_gate = bool(ctx.force_pre_swap_fire_gate)
    on_player_lethal = ctx.on_player_lethal

    weapon_id = player.weapon.weapon_id
    weapon = weapon_entry(weapon_id)

    if (not force_pre_swap_fire_gate) and player.weapon.shot_cooldown > 0.0:
        return WeaponFireResult(fired=False)
    if not input_state.fire_down:
        return WeaponFireResult(fired=False)

    ammo_cost = 1.0
    is_fire_bullets = float(player.fire_bullets_timer) > 0.0
    if (not force_pre_swap_fire_gate) and player.weapon.reload_timer > 0.0:
        if player.experience <= 0:
            return WeaponFireResult(fired=False)
        if perk_active(player, PerkId.REGRESSION_BULLETS):
            ammo_class = int(weapon.ammo_class) if weapon.ammo_class is not None else 0

            reload_time = float(weapon.reload_time)
            factor = 4.0 if ammo_class == 1 else 200.0
            player.experience = int(float(player.experience) - reload_time * factor)
            if player.experience < 0:
                player.experience = 0
        elif perk_active(player, PerkId.AMMUNITION_WITHIN):
            ammo_class = int(weapon.ammo_class) if weapon.ammo_class is not None else 0

            from ..player_damage import player_take_damage

            cost = 0.15 if ammo_class == 1 else 1.0
            player_take_damage(
                state,
                player,
                cost,
                dt=dt,
                rand=state.rng.rand,
                players=players,
                on_lethal=(lambda: on_player_lethal(player)) if on_player_lethal is not None else None,
            )
        else:
            return WeaponFireResult(fired=False)

    pellet_count = int(weapon.pellet_count)
    fire_bullets_weapon = weapon_entry_for_projectile_type_id(ProjectileTemplateId.FIRE_BULLETS)

    shot_cooldown = float(f32(float(weapon.shot_cooldown)))
    weapon_spread_heat = float(weapon.spread_heat_inc)
    fire_bullets_spread_heat = float(fire_bullets_weapon.spread_heat_inc)

    if is_fire_bullets and pellet_count == 1:
        shot_cooldown = float(f32(float(fire_bullets_weapon.shot_cooldown)))

    spread_heat_base = fire_bullets_spread_heat if is_fire_bullets else weapon_spread_heat
    spread_inc = spread_heat_base * 1.3

    if perk_active(player, PerkId.FASTSHOT):
        shot_cooldown = float(f32(float(shot_cooldown) * 0.88))
    if perk_active(player, PerkId.SHARPSHOOTER):
        shot_cooldown = float(f32(float(shot_cooldown) * 1.05))
    player.weapon.shot_cooldown = max(0.0, float(f32(float(shot_cooldown))))

    aim = input_state.aim
    aim_delta = aim - player.pos
    aim_heading = float(heading_from_delta_f32(dx=float(aim_delta.x), dy=float(aim_delta.y)))

    muzzle = player.pos + Vec2.from_heading(aim_heading).rotated(-0.150915) * 16.0
    state.effects.spawn_shell_casing(
        pos=muzzle,
        aim_heading=aim_heading,
        weapon_flags=int(weapon.flags or 0),
        rand=state.rng.rand,
        detail_preset=detail_preset,
    )

    shot_angle = _native_shot_angle_with_jitter(
        aim=aim,
        player_pos=player.pos,
        spread_heat=float(player.spread_heat),
        rand=state.rng.rand,
    )
    particle_angle = Vec2.from_heading(shot_angle).to_angle()
    if weapon_id in (WeaponId.FLAMETHROWER, WeaponId.BLOW_TORCH, WeaponId.HR_FLAMER):
        particle_angle = Vec2.from_heading(aim_heading).to_angle()

    # Native `player_fire_weapon` consumes one RNG draw for shot SFX variant
    # selection on every non-Fire-Bullets shot.
    if not is_fire_bullets:
        state.rng.rand()

    owner = owner_ref_for_player(player.index)
    projectile_owner = owner_ref_for_player_projectiles(state, player.index)
    shot_count = 1
    spawn_muzzle_after_projectile = bool(is_fire_bullets) or int(weapon_id) in _NATIVE_FIRE_MUZZLE_AFTER_PROJECTILE
    if not spawn_muzzle_after_projectile:
        _spawn_native_fire_muzzle_sprites(
            state=state,
            weapon_id=int(weapon_id),
            muzzle=muzzle,
            aim_heading=float(aim_heading),
            fire_bullets_active=bool(is_fire_bullets),
        )

    recipe = resolve_fire_recipe(
        weapon_id=weapon_id,
        pellet_count=pellet_count,
        fire_bullets_active=is_fire_bullets,
    )
    ammo_cost = float(recipe.ammo_cost)

    match recipe.mode:
        case PrimaryPelletsMode(type_id=type_id, count=count, jitter=jitter_rule, speed_scale=speed_rule):
            if type_id is None:
                raise ValueError(f"missing projectile type in recipe for weapon {int(weapon_id)}")
            pellets = max(0, int(count if count is not None else 0))
            shot_count = pellets
            meta = travel_budget_for_type_id(type_id)
            for _ in range(pellets):
                angle = _apply_pellet_jitter(
                    shot_angle=float(shot_angle),
                    rand=state.rng.rand,
                    jitter_rule=jitter_rule,
                )
                proj_id = state.projectiles.spawn(
                    pos=muzzle,
                    angle=angle,
                    type_id=type_id,
                    owner=projectile_owner,
                    travel_budget=meta,
                )
                _apply_speed_scale_rule(
                    state=state,
                    proj_id=int(proj_id),
                    speed_rule=speed_rule,
                )
        case SecondaryShotMode(type_id=type_id, targeting=targeting):
            target_hint = None
            spawn_creatures = None
            match targeting:
                case UseAimTargetHint():
                    target_hint = aim
                    spawn_creatures = creatures
                case _:
                    pass
            state.secondary_projectiles.spawn_from_spec(
                SecondarySpawnSpec(
                    pos=muzzle,
                    angle=shot_angle,
                    type_id=type_id,
                    owner=owner,
                    target_hint=target_hint,
                    creatures=spawn_creatures,
                ),
            )
        case ParticleStreamMode(style=style, slow=slow):
            if slow:
                state.particles.spawn_particle_slow(
                    pos=muzzle,
                    angle=Vec2.from_heading(shot_angle).to_angle(),
                    owner=owner,
                )
            else:
                particle_id = state.particles.spawn_particle(
                    pos=muzzle,
                    angle=particle_angle,
                    intensity=1.0,
                    owner=owner,
                )
                if style is not None:
                    state.particles.entries[particle_id].style_id = style
        case MultiPlasmaFanMode():
            # Multi-Plasma: 5-shot fixed spread using type 0x09 and 0x0B.
            shot_count = 5
            spread_small = math.pi / 10
            spread_large = math.pi / 6
            patterns: tuple[tuple[float, ProjectileTemplateId], ...] = (
                (-spread_small, ProjectileTemplateId.PLASMA_RIFLE),
                (-spread_large, ProjectileTemplateId.PLASMA_MINIGUN),
                (0.0, ProjectileTemplateId.PLASMA_RIFLE),
                (spread_large, ProjectileTemplateId.PLASMA_MINIGUN),
                (spread_small, ProjectileTemplateId.PLASMA_RIFLE),
            )
            for angle_offset, type_id in patterns:
                state.projectiles.spawn(
                    pos=muzzle,
                    angle=shot_angle + angle_offset,
                    type_id=type_id,
                    owner=projectile_owner,
                    travel_budget=travel_budget_for_type_id(type_id),
                )
        case SwarmerDumpMode():
            # Mini-Rocket Swarmers -> secondary type 2 (fires the full clip in a spread).
            rocket_count = max(1, int(player.weapon.ammo))
            if bool(state.preserve_bugs):
                # Native bug: step scales by ammo (`ammo * pi/3`), which aliases to identical headings
                # for some clip sizes (e.g. 6 rockets), causing visible clumping.
                step = float(rocket_count) * (math.pi / 3.0)
                angle = (shot_angle - math.pi) - step * float(rocket_count) * 0.5
            else:
                spread = math.pi * (2.0 / 3.0)
                step = 0.0 if rocket_count <= 1 else spread / float(rocket_count - 1)
                angle = shot_angle - spread * 0.5
            for _ in range(rocket_count):
                state.secondary_projectiles.spawn_from_spec(
                    SecondarySpawnSpec(
                        pos=muzzle,
                        angle=angle,
                        type_id=SecondaryProjectileTypeId.HOMING_ROCKET,
                        owner=owner,
                        target_hint=aim,
                        creatures=creatures,
                    ),
                )
                angle += step
            ammo_cost = float(rocket_count)
            shot_count = rocket_count

    if 0 <= int(player.index) < len(state.shots_fired):
        state.shots_fired[int(player.index)] += int(shot_count)
        if 0 <= weapon_id < WEAPON_COUNT_SIZE:
            state.weapon_shots_fired[int(player.index)][weapon_id] += int(shot_count)

    if spawn_muzzle_after_projectile:
        _spawn_native_fire_muzzle_sprites(
            state=state,
            weapon_id=int(weapon_id),
            muzzle=muzzle,
            aim_heading=float(aim_heading),
            fire_bullets_active=bool(is_fire_bullets),
        )

    if not perk_active(player, PerkId.SHARPSHOOTER):
        player.spread_heat = min(0.48, max(0.0, player.spread_heat + spread_inc))

    muzzle_inc = weapon_spread_heat
    if is_fire_bullets and pellet_count == 1:
        muzzle_inc = fire_bullets_spread_heat
    player.muzzle_flash_alpha = min(1.0, player.muzzle_flash_alpha)
    player.muzzle_flash_alpha = min(1.0, player.muzzle_flash_alpha + muzzle_inc)
    player.muzzle_flash_alpha = min(0.8, player.muzzle_flash_alpha)

    player.shot_seq += 1
    if state.bonuses.reflex_boost <= 0.0 and not is_fire_bullets:
        # Native allows ammo to cross below zero for reload-time firing paths
        # (for example Regression Bullets), and replay checkpoints rely on that.
        player.weapon.ammo = float(player.weapon.ammo) - float(ammo_cost)
    reload_start_gate_open = bool(player.weapon.reload_timer <= 0.0)
    if force_pre_swap_fire_gate:
        # Alt-weapon same-tick fire uses the pre-swap gate (reload_timer==0) for
        # reload restart eligibility after ammo drains below zero.
        reload_start_gate_open = True
    if player.weapon.ammo <= 0.0 and reload_start_gate_open:
        player_start_reload(player, state)
    return WeaponFireResult(fired=True, shot_count=int(shot_count), ammo_cost=float(ammo_cost))



from __future__ import annotations

from collections import Counter
from typing import Literal, TypeAlias

import msgspec

from ...bonuses.ids import BonusId, bonus_display_name
from ...game_modes import GameMode
from ...perks.ids import PerkId, perk_display_name
from ...replay import Replay
from ...weapons import WeaponId, weapon_display_name
from ..hooks import TickResult
from ..input_providers import GameCommand, PerkMenuOpenCommand
from ..state_types import BonusPickupEvent, PlayerState
from .playback_driver import PlaybackDriver, PlaybackWalkHooks
from .setup import ReplayRunnerError

_EPSILON = 1e-6
ReplayInfoCoreEventKind = Literal[
    "bonus_pickup",
    "weapon_change",
    "perk_pick",
    "level_up",
    "health_damage",
    "health_heal",
    "player_death",
]
ReplayInfoExtraEventKind = Literal[
    "creature_deaths",
    "perk_menu_open",
]
ReplayInfoEventKind: TypeAlias = ReplayInfoCoreEventKind | ReplayInfoExtraEventKind

_CORE_EVENT_KINDS: frozenset[ReplayInfoCoreEventKind] = frozenset(
    (
        "bonus_pickup",
        "weapon_change",
        "perk_pick",
        "level_up",
        "health_damage",
        "health_heal",
        "player_death",
    ),
)


class ReplayInfoTimelineEvent(msgspec.Struct, frozen=True):
    tick_index: int
    elapsed_ms: int
    kind: ReplayInfoEventKind
    player_index: int | None
    detail: str
    data: dict[str, object]


class ReplayInfoResult(msgspec.Struct, frozen=True):
    game_mode_id: GameMode
    tick_rate: int
    ticks_simulated: int
    elapsed_ms: int
    player_count: int
    timeline: list[ReplayInfoTimelineEvent]


class _PlayerSnapshot(msgspec.Struct, frozen=True):
    health: float
    level: int
    experience: int
    weapon_id: WeaponId
    perk_counts: tuple[int, ...]


def _capture_snapshots(players: list[PlayerState]) -> list[_PlayerSnapshot]:
    snapshots: list[_PlayerSnapshot] = []
    for player in players:
        snapshots.append(
            _PlayerSnapshot(
                health=player.health,
                level=player.level,
                experience=player.experience,
                weapon_id=player.weapon.weapon_id,
                perk_counts=tuple(player.perk_counts),
            ),
        )
    return snapshots


def _append_event(
    timeline: list[ReplayInfoTimelineEvent],
    *,
    tick_index: int,
    elapsed_ms: int,
    kind: ReplayInfoEventKind,
    player_index: int | None,
    detail: str,
    data: dict[str, object],
    player_filter: int | None,
    include_extra_events: bool,
) -> None:
    if kind not in _CORE_EVENT_KINDS and not include_extra_events:
        return
    if player_filter is not None and player_index is not None and player_index != player_filter:
        return
    timeline.append(
        ReplayInfoTimelineEvent(
            tick_index=tick_index,
            elapsed_ms=elapsed_ms,
            kind=kind,
            player_index=player_index,
            detail=detail,
            data=data,
        ),
    )


def _append_extra_replay_commands(
    *,
    commands: list[GameCommand],
    tick_index: int,
    elapsed_ms: int,
    timeline: list[ReplayInfoTimelineEvent],
    player_filter: int | None,
    include_extra_events: bool,
) -> None:
    if not include_extra_events:
        return
    for cmd in commands:
        if isinstance(cmd, PerkMenuOpenCommand):
            _append_event(
                timeline,
                tick_index=tick_index,
                elapsed_ms=elapsed_ms,
                kind="perk_menu_open",
                player_index=cmd.player_index,
                detail=f"p{cmd.player_index} perk menu opened",
                data={"player_index": cmd.player_index},
                player_filter=player_filter,
                include_extra_events=True,
            )


def _append_bonus_pickup_events(
    *,
    tick_index: int,
    elapsed_ms: int,
    timeline: list[ReplayInfoTimelineEvent],
    pickups: list[BonusPickupEvent],
    preserve_bugs: bool,
    player_filter: int | None,
    include_extra_events: bool,
) -> None:
    for pickup in pickups:
        bonus_id = pickup.bonus_id
        bonus_name = bonus_display_name(bonus_id, preserve_bugs=preserve_bugs)
        detail = f"p{pickup.player_index} picked {bonus_name} ({bonus_id}) amount={pickup.amount}"
        data: dict[str, object] = {
            "bonus_id": bonus_id,
            "bonus_name": bonus_name,
            "amount": pickup.amount,
        }
        if bonus_id == BonusId.WEAPON:
            weapon_id = WeaponId(pickup.amount)
            weapon_name = weapon_display_name(weapon_id, preserve_bugs=preserve_bugs)
            detail += f" -> {weapon_name}"
            data["weapon_id"] = weapon_id
            data["weapon_name"] = weapon_name
        _append_event(
            timeline,
            tick_index=tick_index,
            elapsed_ms=elapsed_ms,
            kind="bonus_pickup",
            player_index=pickup.player_index,
            detail=detail,
            data=data,
            player_filter=player_filter,
            include_extra_events=include_extra_events,
        )


def _append_snapshot_diff_events(
    *,
    tick_index: int,
    elapsed_ms: int,
    before: list[_PlayerSnapshot],
    after: list[_PlayerSnapshot],
    timeline: list[ReplayInfoTimelineEvent],
    preserve_bugs: bool,
    gore_disabled: int,
    player_filter: int | None,
    include_extra_events: bool,
) -> None:
    players_len = min(len(before), len(after))
    for idx in range(players_len):
        pre = before[idx]
        post = after[idx]

        if pre.weapon_id != post.weapon_id:
            weapon_before_name = weapon_display_name(pre.weapon_id, preserve_bugs=preserve_bugs)
            weapon_after_name = weapon_display_name(post.weapon_id, preserve_bugs=preserve_bugs)
            _append_event(
                timeline,
                tick_index=tick_index,
                elapsed_ms=elapsed_ms,
                kind="weapon_change",
                player_index=idx,
                detail=f"p{idx} weapon {weapon_before_name} -> {weapon_after_name}",
                data={
                    "weapon_id_before": pre.weapon_id,
                    "weapon_name_before": weapon_before_name,
                    "weapon_id_after": post.weapon_id,
                    "weapon_name_after": weapon_after_name,
                },
                player_filter=player_filter,
                include_extra_events=include_extra_events,
            )

        if post.level > pre.level:
            _append_event(
                timeline,
                tick_index=tick_index,
                elapsed_ms=elapsed_ms,
                kind="level_up",
                player_index=idx,
                detail=f"p{idx} level {pre.level} -> {post.level} (xp={post.experience})",
                data={
                    "level_before": pre.level,
                    "level_after": post.level,
                    "xp": post.experience,
                },
                player_filter=player_filter,
                include_extra_events=include_extra_events,
            )

        perk_len = min(len(pre.perk_counts), len(post.perk_counts))
        for perk_id in range(perk_len):
            before_count = pre.perk_counts[perk_id]
            after_count = post.perk_counts[perk_id]
            if after_count <= before_count:
                continue
            perk_name = perk_display_name(
                PerkId(perk_id),
                gore_disabled=gore_disabled,
                preserve_bugs=preserve_bugs,
            )
            _append_event(
                timeline,
                tick_index=tick_index,
                elapsed_ms=elapsed_ms,
                kind="perk_pick",
                player_index=idx,
                detail=f"p{idx} perk {perk_name} ({perk_id}) x{after_count}",
                data={
                    "perk_id": perk_id,
                    "perk_name": perk_name,
                    "count_before": before_count,
                    "count_after": after_count,
                },
                player_filter=player_filter,
                include_extra_events=include_extra_events,
            )

        health_before = pre.health
        health_after = post.health
        if health_after < health_before - _EPSILON:
            amount = health_before - health_after
            _append_event(
                timeline,
                tick_index=tick_index,
                elapsed_ms=elapsed_ms,
                kind="health_damage",
                player_index=idx,
                detail=f"p{idx} damage {amount:.6f} (health {health_before:.6f}->{health_after:.6f})",
                data={
                    "amount": amount,
                    "health_before": health_before,
                    "health_after": health_after,
                },
                player_filter=player_filter,
                include_extra_events=include_extra_events,
            )
        elif health_after > health_before + _EPSILON:
            amount = health_after - health_before
            _append_event(
                timeline,
                tick_index=tick_index,
                elapsed_ms=elapsed_ms,
                kind="health_heal",
                player_index=idx,
                detail=f"p{idx} heal {amount:.6f} (health {health_before:.6f}->{health_after:.6f})",
                data={
                    "amount": amount,
                    "health_before": health_before,
                    "health_after": health_after,
                },
                player_filter=player_filter,
                include_extra_events=include_extra_events,
            )

        if health_before > 0.0 and health_after <= 0.0:
            _append_event(
                timeline,
                tick_index=tick_index,
                elapsed_ms=elapsed_ms,
                kind="player_death",
                player_index=idx,
                detail=f"p{idx} died (health {health_before:.6f}->{health_after:.6f})",
                data={
                    "health_before": health_before,
                    "health_after": health_after,
                },
                player_filter=player_filter,
                include_extra_events=include_extra_events,
            )


def _validate_player_filter(*, replay: Replay, player_index: int | None) -> int | None:
    if player_index is None:
        return None
    if player_index < 0:
        raise ReplayRunnerError(f"invalid player_index filter: {player_index}")
    if replay.header.player_count > 0 and player_index >= replay.header.player_count:
        raise ReplayRunnerError(
            f"player_index filter out of range: {player_index} (player_count={replay.header.player_count})",
        )
    return player_index


def collect_replay_info(
    driver: PlaybackDriver,
    *,
    player_index: int | None = None,
    include_extra_events: bool = True,
) -> ReplayInfoResult:
    replay = driver.replay
    mode = driver.mode_id
    player_filter = _validate_player_filter(replay=replay, player_index=player_index)
    timeline: list[ReplayInfoTimelineEvent] = []
    before: list[_PlayerSnapshot] | None = None

    def _append_tick(tick_result: TickResult, *, after_players: list[PlayerState], before: list[_PlayerSnapshot]) -> None:
        source_tick = tick_result.source_tick
        tick = tick_result.payload
        after = _capture_snapshots(after_players)

        elapsed_ms = int(tick.elapsed_ms)
        if mode != GameMode.RUSH:
            _append_extra_replay_commands(
                commands=source_tick.commands,
                tick_index=int(source_tick.tick_index),
                elapsed_ms=elapsed_ms,
                timeline=timeline,
                player_filter=player_filter,
                include_extra_events=include_extra_events,
            )

        _append_bonus_pickup_events(
            tick_index=int(source_tick.tick_index),
            elapsed_ms=elapsed_ms,
            timeline=timeline,
            pickups=tick.step.events.pickups,
            preserve_bugs=replay.header.preserve_bugs,
            player_filter=player_filter,
            include_extra_events=include_extra_events,
        )

        if tick.step.events.deaths:
            _append_event(
                timeline,
                tick_index=int(source_tick.tick_index),
                elapsed_ms=elapsed_ms,
                kind="creature_deaths",
                player_index=None,
                detail=f"creature deaths={len(tick.step.events.deaths)}",
                data={"count": len(tick.step.events.deaths)},
                player_filter=player_filter,
                include_extra_events=include_extra_events,
            )

        _append_snapshot_diff_events(
            tick_index=int(source_tick.tick_index),
            elapsed_ms=elapsed_ms,
            before=before,
            after=after,
            timeline=timeline,
            preserve_bugs=replay.header.preserve_bugs,
            gore_disabled=replay.header.gore_disabled,
            player_filter=player_filter,
            include_extra_events=include_extra_events,
        )

    def _before_tick(_tick_index: int, world, _dt_tick: float) -> None:
        nonlocal before
        before = _capture_snapshots(world.players)

    def _after_tick(tick_result: TickResult, world) -> None:
        assert before is not None, "missing pre-step replay snapshot"
        _append_tick(tick_result, after_players=world.players, before=before)

    walk_result = driver.walk_ticks(
        hooks=PlaybackWalkHooks(
            before_tick=_before_tick,
            after_tick=_after_tick,
        ),
    )
    run_result = driver.build_run_result(ticks=int(walk_result.ticks_completed))

    return ReplayInfoResult(
        game_mode_id=mode,
        tick_rate=replay.header.tick_rate,
        ticks_simulated=int(walk_result.ticks_completed),
        elapsed_ms=int(run_result.elapsed_ms),
        player_count=len(driver.world.players),
        timeline=timeline,
    )


def event_counts_by_kind(timeline: list[ReplayInfoTimelineEvent]) -> dict[str, int]:
    counts: Counter[ReplayInfoEventKind] = Counter()
    for event in timeline:
        counts[event.kind] += 1
    return {str(kind): int(count) for kind, count in sorted(counts.items())}



from __future__ import annotations

from typing import cast

import msgspec

from ..owner_ref import OwnerRef
from ..sim.timing import ftol_ms_i32


class SnapshotVec2(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    x: float
    y: float


class SnapshotWeapon(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    weapon_id: int
    ammo: float
    clip_size: int
    reload_active: bool
    reload_timer: float
    reload_timer_max: float
    shot_cooldown: float


class SnapshotPlayer(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    index: int
    pos: SnapshotVec2
    health: float
    weapon: SnapshotWeapon
    experience: int
    level: int


class SnapshotStatus(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    quest_unlock_index: int
    quest_unlock_index_full: int
    weapon_usage_counts: list[int]


class SnapshotBonusTimers(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    weapon_power_up_ms: int
    reflex_boost_ms: int
    energizer_ms: int
    double_experience_ms: int
    freeze_ms: int


class SnapshotGameplay(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    mode_id: int
    quest_stage_major: int
    quest_stage_minor: int
    perk_pending_count: int
    perk_choices_dirty: bool
    bonus_timers: SnapshotBonusTimers
    status: SnapshotStatus


class SimStateSnapshot(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    gameplay: SnapshotGameplay
    players: list[SnapshotPlayer]


class RngStreamRow(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    tick_call_index: int
    value_15: int
    state_before_u32: int
    state_after_u32: int
    caller_static: str | None = None
    branch_id: str | None = None


class TimingSampleRow(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    tick_index: int
    gameplay_frame: int | None = None
    phase: str = ""
    write_kind: str = "snapshot"
    frame_dt_f32: float | None = None
    frame_dt_ms_i32: int | None = None
    frame_dt_ms_f32: float | None = None
    time_scale_active_entry: bool | None = None
    time_scale_active_current: bool | None = None
    time_scale_factor: float | None = None
    bonus_reflex_boost_timer: float | None = None
    mode_fn: str | None = None
    player_index: int | None = None


class CreatureEntitySample(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    uid: int
    generation: int
    pool_kind: str
    index: int
    active: bool
    type_id: int
    hp: float
    pos: SnapshotVec2
    flags: int
    ai_mode: int
    link_index: int
    heading: float
    target_heading: float
    orbit_angle: float
    orbit_radius: float
    lifecycle_stage: float


class ProjectileEntitySample(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    uid: int
    generation: int
    pool_kind: str
    index: int
    active: bool
    type_id: int
    angle: float
    pos: SnapshotVec2
    vel: SnapshotVec2
    life_timer: float
    speed_scale: float
    damage_pool: float
    hit_radius: float
    travel_budget: float
    owner: OwnerRef


class SecondaryProjectileEntitySample(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    uid: int
    generation: int
    pool_kind: str
    index: int
    active: bool
    type_id: int
    angle: float
    pos: SnapshotVec2
    vel: SnapshotVec2
    speed: float
    trail_timer: float
    owner: OwnerRef
    target_id: int


class BonusEntitySample(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    uid: int
    generation: int
    pool_kind: str
    index: int
    active: bool
    bonus_id: int
    picked: bool
    time_left: float
    time_max: float
    pos: SnapshotVec2
    amount: int


class EntitySamplesSnapshot(msgspec.Struct, frozen=True, forbid_unknown_fields=True):
    creatures: list[CreatureEntitySample]
    projectiles: list[ProjectileEntitySample]
    secondary_projectiles: list[SecondaryProjectileEntitySample]
    bonuses: list[BonusEntitySample]


def bonus_timer_ms(value: float) -> int:
    return max(0, int(ftol_ms_i32(float(value))))


def _to_builtin(value: object) -> object:
    return msgspec.to_builtins(value)


def validate_sim_state(value: object, *, field: str) -> dict[str, object]:
    try:
        validated = msgspec.convert(value, type=SimStateSnapshot)
    except (msgspec.ValidationError, TypeError, ValueError) as exc:
        raise ValueError(f"{field} must be a valid SimStateSnapshot payload") from exc
    return cast("dict[str, object]", _to_builtin(validated))


def validate_rng_stream(value: object, *, field: str) -> list[object]:
    try:
        validated = msgspec.convert(value, type=list[RngStreamRow])
    except (msgspec.ValidationError, TypeError, ValueError) as exc:
        raise ValueError(f"{field} must be a valid rng_stream payload") from exc
    return cast("list[object]", _to_builtin(validated))


def validate_entity_samples(value: object, *, field: str) -> dict[str, object]:
    try:
        validated = msgspec.convert(value, type=EntitySamplesSnapshot)
    except (msgspec.ValidationError, TypeError, ValueError) as exc:
        raise ValueError(f"{field} must be a valid EntitySamplesSnapshot payload") from exc
    return cast("dict[str, object]", _to_builtin(validated))


def validate_timing_samples(value: object, *, field: str) -> list[object]:
    try:
        validated = msgspec.convert(value, type=list[TimingSampleRow])
    except (msgspec.ValidationError, TypeError, ValueError) as exc:
        raise ValueError(f"{field} must be a valid timing_samples payload") from exc
    return cast("list[object]", _to_builtin(validated))



from __future__ import annotations

from .weapons import WeaponId

# Save-status stores 53 u32 entries:
# - slot 0 is unused
# - tracked weapon ids map 1:1 to slots 1..52
# - weapon id 53 has no safe slot in this table
WEAPON_USAGE_SLOT_COUNT = 53
WEAPON_USAGE_TRACKED_WEAPON_ID_MIN = int(WeaponId.PISTOL)
WEAPON_USAGE_TRACKED_WEAPON_ID_MAX = WEAPON_USAGE_SLOT_COUNT - 1


def weapon_usage_slot_for_weapon_id(weapon_id: int) -> int | None:
    """Map a weapon id to save-status usage slot, or `None` if untracked."""

    weapon_id = int(weapon_id)
    if WEAPON_USAGE_TRACKED_WEAPON_ID_MIN <= weapon_id <= WEAPON_USAGE_TRACKED_WEAPON_ID_MAX:
        return weapon_id
    return None



from __future__ import annotations

from collections.abc import Callable, MutableSequence
from enum import IntEnum
from typing import Protocol

import msgspec

from grim.color import RGBA
from grim.geom import Vec2
from grim.rand import CrandLike

from ..effects import EffectPool
from ..owner_ref import OwnerRef


class _BonusesLike(Protocol):
    freeze: float


class _SpriteEffectsLike(Protocol):
    def spawn(self, *, pos: Vec2, vel: Vec2, scale: float = 1.0, color: RGBA | None = None) -> int: ...


class ProjectileRuntimeState(Protocol):
    bonus_spawn_guard: bool
    camera_shake_pulses: int
    shock_chain_links_left: int
    shock_chain_projectile_id: int

    @property
    def effects(self) -> EffectPool: ...

    @property
    def sprite_effects(self) -> _SpriteEffectsLike: ...

    @property
    def rng(self) -> CrandLike: ...

    @property
    def bonuses(self) -> _BonusesLike: ...

    @property
    def sfx_queue(self) -> MutableSequence[str]: ...

    @property
    def shots_hit(self) -> MutableSequence[int]: ...


class FxQueueLike(Protocol):
    def add(
        self,
        *,
        effect_id: int,
        pos: Vec2,
        width: float,
        height: float,
        rotation: float,
        rgba: RGBA,
    ) -> bool: ...

    def add_random(self, *, pos: Vec2, rand: Callable[[], int]) -> bool: ...


MAIN_PROJECTILE_POOL_SIZE = 0x60
SECONDARY_PROJECTILE_POOL_SIZE = 0x40


class ProjectileTemplateId(IntEnum):
    # Values are projectile type ids (not weapon ids). Based on the decompile
    # for `player_fire_weapon` and `projectile_update`.
    PISTOL = 0x01
    ASSAULT_RIFLE = 0x02
    SHOTGUN = 0x03
    SUBMACHINE_GUN = 0x05
    GAUSS_GUN = 0x06
    PLASMA_RIFLE = 0x09
    PLASMA_MINIGUN = 0x0B
    PULSE_GUN = 0x13
    ION_RIFLE = 0x15
    ION_MINIGUN = 0x16
    ION_CANNON = 0x17
    SHRINKIFIER = 0x18
    BLADE_GUN = 0x19
    SPIDER_PLASMA = 0x1A
    PLASMA_CANNON = 0x1C
    SPLITTER_GUN = 0x1D
    PLAGUE_SPREADER = 0x29
    RAINBOW_GUN = 0x2B
    FIRE_BULLETS = 0x2D


class SecondaryProjectileTypeId(IntEnum):
    NONE = 0
    ROCKET = 1
    HOMING_ROCKET = 2
    DETONATION = 3
    ROCKET_MINIGUN = 4


def _rng_zero() -> int:
    return 0


class CreatureDamageApplier(Protocol):
    def __call__(self, creature_index: int, damage: float, damage_type: int, impulse: Vec2, owner: OwnerRef, /) -> None: ...


SecondaryDetonationKillHandler = Callable[[int], None]


class ProjectileCollisionProfile(msgspec.Struct, frozen=True):
    hit_radius: float
    initial_damage_pool: float


class ProjectileHit(msgspec.Struct, frozen=True):
    type_id: ProjectileTemplateId
    origin: Vec2
    hit: Vec2
    target: Vec2


class Projectile(msgspec.Struct):
    active: bool = False
    angle: float = 0.0
    pos: Vec2 = Vec2()
    origin: Vec2 = Vec2()
    vel: Vec2 = Vec2()
    type_id: ProjectileTemplateId = ProjectileTemplateId.PISTOL
    life_timer: float = 0.0
    reserved: float = 0.0
    speed_scale: float = 1.0
    damage_pool: float = 1.0
    hit_radius: float = 1.0
    travel_budget: float = 0.0
    owner: OwnerRef = msgspec.field(default_factory=OwnerRef.none)
    hits_players: bool = False


class SecondaryProjectile(msgspec.Struct):
    active: bool = False
    angle: float = 0.0
    speed: float = 0.0
    pos: Vec2 = Vec2()
    vel: Vec2 = Vec2()
    detonation_t: float = 0.0
    detonation_scale: float = 1.0
    type_id: SecondaryProjectileTypeId = SecondaryProjectileTypeId.NONE
    owner: OwnerRef = msgspec.field(default_factory=lambda: OwnerRef.from_local_player(0))
    trail_timer: float = 0.0
    target_id: int = -1


__all__ = [
    "CreatureDamageApplier",
    "FxQueueLike",
    "MAIN_PROJECTILE_POOL_SIZE",
    "Projectile",
    "ProjectileCollisionProfile",
    "ProjectileHit",
    "ProjectileRuntimeState",
    "ProjectileTemplateId",
    "OwnerRef",
    "SecondaryDetonationKillHandler",
    "SECONDARY_PROJECTILE_POOL_SIZE",
    "SecondaryProjectile",
    "SecondaryProjectileTypeId",
    "_rng_zero",
    "_SpriteEffectsLike",
]



from __future__ import annotations

import html
from pathlib import Path
from typing import cast

from .channel_helpers import ENTITY_SAMPLE_KINDS, checkpoint_channel, entity_samples_channel
from .diff import diff_traces
from .schema import TickRecord
from .trace import TraceReader


def _entity_counts(row: TickRecord | None) -> dict[str, int]:
    if row is None:
        return {kind: 0 for kind in ENTITY_SAMPLE_KINDS}
    samples = entity_samples_channel(row)
    if samples is None:
        return {kind: 0 for kind in ENTITY_SAMPLE_KINDS}
    return {
        "creatures": len(samples.creatures),
        "projectiles": len(samples.projectiles),
        "secondary_projectiles": len(samples.secondary_projectiles),
        "bonuses": len(samples.bonuses),
    }


def _checkpoint_value(row: TickRecord | None, key: str) -> object:
    if row is None:
        return None
    checkpoint = checkpoint_channel(row)
    if checkpoint is None:
        return None
    match str(key):
        case "score_xp":
            return int(checkpoint.score_xp)
        case "kills":
            return int(checkpoint.kills)
        case _:
            return None


def _score_value(value: object) -> int:
    match value:
        case int() as score:
            return int(score)
        case _:
            return 0


def _timeline_rows(
    *,
    golden_trace: Path,
    candidate_trace: Path,
    tick_start: int,
    tick_end: int,
) -> list[dict[str, object]]:
    with TraceReader(Path(golden_trace)) as expected, TraceReader(Path(candidate_trace)) as candidate:
        expected_rows = {
            int(row.tick_index): row
            for row in expected.iter_ticks(tick_start=tick_start, tick_end=tick_end)
        }
        candidate_rows = {
            int(row.tick_index): row
            for row in candidate.iter_ticks(tick_start=tick_start, tick_end=tick_end)
        }

    all_ticks = sorted(set(expected_rows) | set(candidate_rows))
    out: list[dict[str, object]] = []
    for tick in all_ticks:
        expected_row = expected_rows.get(tick)
        candidate_row = candidate_rows.get(tick)
        expected_counts = _entity_counts(expected_row)
        candidate_counts = _entity_counts(candidate_row)
        expected_score = _score_value(_checkpoint_value(expected_row, "score_xp"))
        candidate_score = _score_value(_checkpoint_value(candidate_row, "score_xp"))
        expected_kills = _score_value(_checkpoint_value(expected_row, "kills"))
        candidate_kills = _score_value(_checkpoint_value(candidate_row, "kills"))
        diverged = bool(
            expected_score != candidate_score
            or expected_kills != candidate_kills
            or expected_counts != candidate_counts,
        )
        out.append(
            {
                "tick_index": int(tick),
                "diverged": diverged,
                "expected_score_xp": int(expected_score),
                "candidate_score_xp": int(candidate_score),
                "expected_kills": int(expected_kills),
                "candidate_kills": int(candidate_kills),
                "expected_entities": expected_counts,
                "candidate_entities": candidate_counts,
            },
        )
    return out


def _default_focus_tick(
    *,
    golden_trace: Path,
    candidate_trace: Path,
) -> int:
    report = diff_traces(
        expected_trace_path=Path(golden_trace),
        actual_trace_path=Path(candidate_trace),
    )
    if report.mismatch is not None:
        return int(report.mismatch.tick_index)
    if report.tick_start is not None:
        return int(report.tick_start)
    with TraceReader(Path(golden_trace)) as trace:
        return int(trace.footer.first_tick)


def _render_html(*, payload: dict[str, object]) -> str:
    rows = cast("list[dict[str, object]]", payload.get("rows", []))
    body_rows: list[str] = []
    focus_tick = payload.get("focus_tick")
    for row in rows:
        tick = row.get("tick_index")
        expected_entities = cast("dict[str, object]", row.get("expected_entities", {}))
        candidate_entities = cast("dict[str, object]", row.get("candidate_entities", {}))
        classes: list[str] = []
        if bool(row.get("diverged")):
            classes.append("diverged")
        if tick == focus_tick:
            classes.append("focus")
        class_attr = f' class="{" ".join(classes)}"' if classes else ""
        tick_attr = f' data-tick="{html.escape(str(tick))}"'
        body_rows.append(
            "<tr"
            + tick_attr
            + class_attr
            + ">"
            + f"<td>{html.escape(str(tick))}</td>"
            + f"<td>{html.escape(str(row.get('expected_score_xp')))}</td>"
            + f"<td>{html.escape(str(row.get('candidate_score_xp')))}</td>"
            + f"<td>{html.escape(str(row.get('expected_kills')))}</td>"
            + f"<td>{html.escape(str(row.get('candidate_kills')))}</td>"
            + f"<td>{html.escape(str(expected_entities.get('creatures')))}</td>"
            + f"<td>{html.escape(str(candidate_entities.get('creatures')))}</td>"
            + f"<td>{html.escape(str(expected_entities.get('projectiles')))}</td>"
            + f"<td>{html.escape(str(candidate_entities.get('projectiles')))}</td>"
            + "</tr>",
        )

    golden_text = html.escape(str(payload.get("golden_trace")))
    candidate_text = html.escape(str(payload.get("candidate_trace")))
    focus_text = html.escape(str(payload.get("focus_tick")))
    start_text = html.escape(str(payload.get("tick_start")))
    end_text = html.escape(str(payload.get("tick_end")))
    rows_html = "\n".join(body_rows)
    return (
        "<!doctype html>\n"
        '<html lang="en">\n'
        "<head>\n"
        '  <meta charset="utf-8" />\n'
        '  <meta name="viewport" content="width=device-width, initial-scale=1" />\n'
        "  <title>Crimson Debug Viz</title>\n"
        "  <style>\n"
        "    :root { --bg: #f4f6f8; --fg: #1d2a35; --muted: #4f5d6b; --line: #d9e1e8; --focus: #ffe39f; --div: #ffd7d7; }\n"
        '    body { margin: 0; font-family: "IBM Plex Sans", "Segoe UI", sans-serif; background: var(--bg); color: var(--fg); }\n'
        "    .wrap { max-width: 1100px; margin: 0 auto; padding: 24px 16px 40px; }\n"
        "    .meta { display: grid; gap: 6px; margin-bottom: 14px; color: var(--muted); font-size: 14px; }\n"
        "    .controls { display: flex; gap: 8px; align-items: center; margin-bottom: 12px; }\n"
        "    .controls button { border: 1px solid var(--line); background: #fff; padding: 6px 10px; cursor: pointer; }\n"
        "    .controls input[type=range] { flex: 1; }\n"
        "    table { width: 100%; border-collapse: collapse; background: #fff; border: 1px solid var(--line); }\n"
        "    th, td { border-bottom: 1px solid var(--line); padding: 8px; text-align: right; font-size: 13px; }\n"
        "    th:first-child, td:first-child { text-align: left; }\n"
        "    tr.diverged { background: var(--div); }\n"
        "    tr.focus { outline: 2px solid #d18f00; outline-offset: -2px; background: var(--focus); }\n"
        "  </style>\n"
        "</head>\n"
        "<body>\n"
        '  <div class="wrap">\n'
        "    <h1>Crimson Debug Viz</h1>\n"
        '    <div class="meta">\n'
        f"      <div>Golden: {golden_text}</div>\n"
        f"      <div>Candidate: {candidate_text}</div>\n"
        f"      <div>Focus tick: {focus_text} | Window: {start_text}..{end_text}</div>\n"
        "    </div>\n"
        '    <div class="controls">\n'
        '      <button id="tick-prev" type="button">Prev</button>\n'
        '      <input id="tick-slider" type="range" min="'
        + start_text
        + '" max="'
        + end_text
        + '" value="'
        + focus_text
        + '" />\n'
        '      <button id="tick-next" type="button">Next</button>\n'
        '      <span id="tick-label">Tick '
        + focus_text
        + "</span>\n"
        "    </div>\n"
        "    <table>\n"
        "      <thead>\n"
        "        <tr>\n"
        "          <th>Tick</th>\n"
        "          <th>G Score</th>\n"
        "          <th>C Score</th>\n"
        "          <th>G Kills</th>\n"
        "          <th>C Kills</th>\n"
        "          <th>G Creatures</th>\n"
        "          <th>C Creatures</th>\n"
        "          <th>G Projectiles</th>\n"
        "          <th>C Projectiles</th>\n"
        "        </tr>\n"
        "      </thead>\n"
        "      <tbody>\n"
        f"{rows_html}\n"
        "      </tbody>\n"
        "    </table>\n"
        "    <script>\n"
        "      (function () {\n"
        "        const slider = document.getElementById('tick-slider');\n"
        "        const label = document.getElementById('tick-label');\n"
        "        const prev = document.getElementById('tick-prev');\n"
        "        const next = document.getElementById('tick-next');\n"
        "        const rows = Array.from(document.querySelectorAll('tbody tr'));\n"
        "        function setTick(tick) {\n"
        "          const n = Number(tick);\n"
        "          slider.value = String(n);\n"
        "          label.textContent = 'Tick ' + String(n);\n"
        "          let focusRow = null;\n"
        "          rows.forEach((row) => {\n"
        "            if (Number(row.getAttribute('data-tick')) === n) {\n"
        "              row.classList.add('focus');\n"
        "              focusRow = row;\n"
        "            } else {\n"
        "              row.classList.remove('focus');\n"
        "            }\n"
        "          });\n"
        "          if (focusRow) {\n"
        "            focusRow.scrollIntoView({ block: 'nearest' });\n"
        "          }\n"
        "        }\n"
        "        slider.addEventListener('input', () => setTick(slider.value));\n"
        "        prev.addEventListener('click', () => setTick(Math.max(Number(slider.min), Number(slider.value) - 1)));\n"
        "        next.addEventListener('click', () => setTick(Math.min(Number(slider.max), Number(slider.value) + 1)));\n"
        "      })();\n"
        "    </script>\n"
        "  </div>\n"
        "</body>\n"
        "</html>\n"
    )


def write_viz_html(
    *,
    golden_trace: Path,
    candidate_trace: Path,
    tick: int | None = None,
    window_before: int = 64,
    window_after: int = 64,
    out_path: Path,
) -> dict[str, object]:
    focus_tick = (
        int(tick)
        if tick is not None
        else _default_focus_tick(
            golden_trace=Path(golden_trace),
            candidate_trace=Path(candidate_trace),
        )
    )
    left = int(focus_tick) - max(0, int(window_before))
    right = int(focus_tick) + max(0, int(window_after))
    rows = _timeline_rows(
        golden_trace=Path(golden_trace),
        candidate_trace=Path(candidate_trace),
        tick_start=left,
        tick_end=right,
    )
    payload: dict[str, object] = {
        "golden_trace": str(golden_trace),
        "candidate_trace": str(candidate_trace),
        "focus_tick": int(focus_tick),
        "tick_start": int(left),
        "tick_end": int(right),
        "row_count": len(rows),
        "rows": rows,
    }
    html_text = _render_html(payload=payload)
    out = Path(out_path)
    out.parent.mkdir(parents=True, exist_ok=True)
    out.write_text(html_text, encoding="utf-8")
    payload["html_path"] = str(out)
    return payload



from __future__ import annotations

import json
from collections.abc import Sequence

import msgspec

from ..replay.checkpoints import ReplayCheckpoint
from .strict_compare import strict_mismatch_payload

DEFAULT_RNG_MARK_ORDER: tuple[str, ...] = (
    "before_world_step",
    "gw_begin",
    "gw_after_weapon_refresh",
    "gw_after_perks_rebuild",
    "gw_after_time_scale",
    "ws_begin",
    "ws_after_perk_effects",
    "ws_after_effects_update",
    "ws_after_creatures",
    "ws_after_projectiles",
    "ws_after_secondary_projectiles",
    "ws_after_particles_update",
    "ws_after_sprite_effects",
    "ws_after_particles",
    "ws_after_player_update_p0",
    "ws_after_player_update",
    "ws_after_bonus_update",
    "ws_after_progression",
    "ws_after_sfx_queue_merge",
    "ws_after_player_damage_sfx",
    "ws_after_sfx",
    "after_world_step",
    "after_stage_spawns",
    "after_wave_spawns",
    "after_rush_spawns",
)

class ReplayDiffFailure(msgspec.Struct, frozen=True):
    kind: str
    tick_index: int
    expected: ReplayCheckpoint
    actual: ReplayCheckpoint | None = None
    first_rng_mark: str | None = None


class ReplayDiffResult(msgspec.Struct, frozen=True):
    ok: bool
    checked_count: int
    first_rng_only_tick: int | None = None
    failure: ReplayDiffFailure | None = None


class CheckpointDeepDiff(msgspec.Struct, frozen=True):
    payload: dict[str, object]
    pretty: str
    diff_count: int


def _checkpoint_to_obj(
    checkpoint: ReplayCheckpoint,
    *,
    include_rng_fields: bool,
) -> dict[str, object]:
    obj = msgspec.to_builtins(checkpoint)
    if not include_rng_fields:
        for key in ("rng_state", "rng_marks"):
            obj.pop(key, None)
    return obj


def checkpoint_deepdiff(
    expected: ReplayCheckpoint,
    actual: ReplayCheckpoint,
) -> CheckpointDeepDiff | None:
    mismatches, diff_count, _pretty = strict_mismatch_payload(expected, actual)
    if int(diff_count) <= 0:
        return None

    payload: dict[str, object] = {"mismatches": mismatches}
    return CheckpointDeepDiff(
        payload=payload,
        pretty=json.dumps(payload, sort_keys=True, indent=2, default=repr),
        diff_count=int(diff_count),
    )


def compare_checkpoints(
    expected: Sequence[ReplayCheckpoint],
    actual: Sequence[ReplayCheckpoint],
    *,
    rng_mark_order: Sequence[str] = DEFAULT_RNG_MARK_ORDER,
) -> ReplayDiffResult:
    actual_by_tick = {int(ckpt.tick_index): ckpt for ckpt in actual}
    first_rng_only_tick: int | None = None
    checked_count = 0

    for exp in expected:
        checked_count += 1
        tick = int(exp.tick_index)
        act = (actual_by_tick[tick] if tick in actual_by_tick else None)
        if act is None:
            return ReplayDiffResult(
                ok=False,
                checked_count=checked_count,
                first_rng_only_tick=first_rng_only_tick,
                failure=ReplayDiffFailure(
                    kind="missing_checkpoint",
                    tick_index=tick,
                    expected=exp,
                    actual=None,
                ),
            )

        if exp == act:
            continue

        exp_no_rng = _checkpoint_to_obj(exp, include_rng_fields=False)
        act_no_rng = _checkpoint_to_obj(act, include_rng_fields=False)
        if exp_no_rng == act_no_rng:
            if first_rng_only_tick is None:
                first_rng_only_tick = tick
            continue

        mark_keys = sorted({*exp.rng_marks.keys(), *act.rng_marks.keys()})
        mark_mismatch: list[str] = []
        for key in mark_keys:
            if key in exp.rng_marks:
                exp_mark = int(exp.rng_marks[key])
            else:
                exp_mark = -1
            if key in act.rng_marks:
                act_mark = int(act.rng_marks[key])
            else:
                act_mark = -1
            if exp_mark != act_mark:
                mark_mismatch.append(key)
        first_mark = next((key for key in rng_mark_order if key in mark_mismatch), mark_mismatch[0] if mark_mismatch else None)
        return ReplayDiffResult(
            ok=False,
            checked_count=checked_count,
            first_rng_only_tick=first_rng_only_tick,
            failure=ReplayDiffFailure(
                kind="state_mismatch",
                tick_index=tick,
                expected=exp,
                actual=act,
                first_rng_mark=first_mark,
            ),
        )

    return ReplayDiffResult(
        ok=True,
        checked_count=checked_count,
        first_rng_only_tick=first_rng_only_tick,
        failure=None,
    )



from __future__ import annotations

from typing import TYPE_CHECKING, TypeAlias

import msgspec

from grim.geom import Vec2

from ..bonuses.ids import BonusId
from ..weapons import WeaponId

PERK_COUNT_SIZE = 0x80


class WeaponSlot(msgspec.Struct):
    weapon_id: WeaponId
    clip_size: int = 0
    ammo: float = 0.0
    reload_active: bool = False
    reload_timer: float = 0.0
    reload_timer_max: float = 0.0
    shot_cooldown: float = 0.0

class PlayerState(msgspec.Struct):
    index: int
    pos: Vec2
    health: float = 100.0
    size: float = 48.0

    speed_multiplier: float = 2.0
    move_speed: float = 0.0
    move_phase: float = 0.0
    heading: float = 0.0
    turn_speed: float = 1.0
    death_timer: float = 16.0
    low_health_timer: float = 100.0

    aim: Vec2 = Vec2()
    aim_heading: float = 0.0
    aim_dir: Vec2 = Vec2(1.0, 0.0)
    evil_eyes_target_creature: int = -1
    auto_target: int = -1

    bonus_aim_hover_index: int = -1
    bonus_aim_hover_timer_ms: float = 0.0

    weapon: WeaponSlot = msgspec.field(default_factory=lambda: WeaponSlot(weapon_id=WeaponId.PISTOL))
    alt_weapon: WeaponSlot | None = None

    shot_seq: int = 0
    weapon_reset_latch: int = 0
    aux_timer: float = 0.0
    spread_heat: float = 0.01
    muzzle_flash_alpha: float = 0.0

    experience: int = 0
    level: int = 1

    perk_counts: list[int] = msgspec.field(default_factory=lambda: [0] * PERK_COUNT_SIZE)
    plaguebearer_active: bool = False
    hot_tempered_timer: float = 0.0
    man_bomb_timer: float = 0.0
    living_fortress_timer: float = 0.0
    fire_cough_timer: float = 0.0

    speed_bonus_timer: float = 0.0
    shield_timer: float = 0.0
    fire_bullets_timer: float = 0.0


class BonusPickupEvent(msgspec.Struct, frozen=True):
    player_index: int
    bonus_id: BonusId
    amount: int
    pos: Vec2


if TYPE_CHECKING:
    from ..gameplay import GameplayState as GameplayState
else:
    GameplayState: TypeAlias = object



from __future__ import annotations

from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import Protocol

from .hooks import TickResult
from .presentation_step import PresentationStepCommands
from .step_pipeline import DeterministicStepResult
from .world_state import WorldEvents


class SimMetadataSink(Protocol):
    def apply_step_metadata(
        self,
        *,
        events: WorldEvents,
        presentation: PresentationStepCommands,
        dt_sim: float,
        game_tune_started: bool,
    ) -> None: ...


@dataclass(frozen=True, slots=True)
class PresentationTickOutput:
    tick_index: int
    dt_sim: float
    presentation: PresentationStepCommands | None


def apply_sim_metadata_tick_result(
    *,
    sim_world: SimMetadataSink,
    tick_result: TickResult,
    game_tune_started: bool,
) -> PresentationTickOutput:
    step = tick_result.payload.step

    apply_tick_to_sim(
        sim_world=sim_world,
        step=step,
        game_tune_started=bool(game_tune_started),
    )
    return PresentationTickOutput(
        tick_index=int(tick_result.source_tick.tick_index),
        dt_sim=float(step.dt_sim),
        presentation=step.presentation,
    )


def apply_tick_to_sim(
    *,
    sim_world: SimMetadataSink,
    step: DeterministicStepResult,
    game_tune_started: bool,
) -> None:
    sim_world.apply_step_metadata(
        events=step.events,
        presentation=step.presentation,
        dt_sim=float(step.dt_sim),
        game_tune_started=bool(game_tune_started),
    )


def apply_sim_metadata_batch(
    *,
    sim_world: SimMetadataSink,
    completed_results: Sequence[TickResult],
    game_tune_started: bool,
) -> list[PresentationTickOutput]:
    return [
        apply_sim_metadata_tick_result(
            sim_world=sim_world,
            tick_result=tick_result,
            game_tune_started=bool(game_tune_started),
        )
        for tick_result in completed_results
    ]


def apply_presentation_outputs(
    *,
    outputs: Sequence[PresentationTickOutput],
    sync_audio_bridge_state: Callable[[], None],
    apply_audio_plan: Callable[[PresentationStepCommands, bool], None],
    update_camera: Callable[[float], None] | None,
    on_output_applied: Callable[[PresentationTickOutput], None] | None = None,
    apply_audio: bool,
) -> None:
    if not outputs:
        return

    sync_audio_bridge_state()
    for output in outputs:
        if output.presentation is None:
            if on_output_applied is not None:
                on_output_applied(output)
            continue
        apply_audio_plan(output.presentation, bool(apply_audio))
        if update_camera is not None:
            update_camera(float(output.dt_sim))
        if on_output_applied is not None:
            on_output_applied(output)



from __future__ import annotations

import math

from grim.geom import Vec2

from ..owner_ref import OwnerRef
from ..projectiles.types import ProjectileTemplateId
from ..sim.state_types import GameplayState, PlayerState
from ..weapons import weapon_entry_for_projectile_type_id


def owner_ref_for_player(player_index: int) -> OwnerRef:
    return OwnerRef.from_player(int(player_index))


def owner_ref_for_player_projectiles(state: GameplayState, player_index: int) -> OwnerRef:
    if not state.friendly_fire_enabled:
        return OwnerRef.from_local_player(0)
    return owner_ref_for_player(player_index)


def travel_budget_for_type_id(type_id: ProjectileTemplateId) -> float:
    return float(weapon_entry_for_projectile_type_id(type_id).travel_budget)


def _resolve_player_slot(players: list[PlayerState], *, player_index: int) -> int | None:
    target_index = int(player_index)
    if 0 <= target_index < len(players):
        direct = players[target_index]
        if int(direct.index) == target_index:
            return int(target_index)
    for slot, player in enumerate(players):
        if int(player.index) == target_index:
            return int(slot)
    return None


def _shots_fired_player_index(
    *,
    state: GameplayState,
    players: list[PlayerState] | None,
    owner: OwnerRef,
    owner_player_index: int | None,
) -> int | None:
    if owner_player_index is not None:
        player_index = int(owner_player_index)
        if 0 <= player_index < len(state.shots_fired):
            return int(player_index)

    if owner.is_player() and not (owner.local_host and owner.index == 0):
        player_index = int(owner.index)
        if 0 <= player_index < len(state.shots_fired):
            return int(player_index)

    if owner.local_host and owner.index == 0 and players and len(players) == 1:
        player_index = int(players[0].index)
        if 0 <= player_index < len(state.shots_fired):
            return int(player_index)

    return None


def _fire_bullets_active(
    players: list[PlayerState] | None,
    *,
    state: GameplayState,
    owner: OwnerRef,
    owner_player_index: int | None,
) -> bool:
    if not players:
        return False

    # Native `projectile_spawn` checks player-1/player-2 Fire Bullets timers
    # globally, regardless of projectile ownership.
    if bool(state.preserve_bugs):
        return any(float(player.fire_bullets_timer) > 0.0 for player in players[:2])

    resolved_owner_slot: int | None = None
    if owner_player_index is not None:
        resolved_owner_slot = _resolve_player_slot(players, player_index=int(owner_player_index))
    elif owner.is_player() and not (owner.local_host and owner.index == 0):
        resolved_owner_slot = _resolve_player_slot(players, player_index=int(owner.index))
    elif owner.local_host and owner.index == 0 and len(players) == 1:
        # Callers that only pass one player are explicitly indicating the owner
        # context (for example OwnerRef.from_local_player(0) with friendly fire disabled).
        resolved_owner_slot = 0

    if resolved_owner_slot is None:
        return False
    if not (0 <= resolved_owner_slot < len(players)):
        return False
    return float(players[resolved_owner_slot].fire_bullets_timer) > 0.0


def projectile_spawn(
    state: GameplayState,
    *,
    players: list[PlayerState] | None,
    pos: Vec2,
    angle: float,
    type_id: ProjectileTemplateId,
    owner: OwnerRef,
    owner_player_index: int | None = None,
    hits_players: bool = False,
) -> int:
    # Mirror `projectile_spawn` (0x00420440) Fire Bullets override.
    if (not state.bonus_spawn_guard) and owner.is_player():
        while True:
            player_index = _shots_fired_player_index(
                state=state,
                players=players,
                owner=owner,
                owner_player_index=owner_player_index,
            )
            state.shots_fired_total += 1
            if player_index is not None:
                state.shots_fired[player_index] += 1
            if type_id == ProjectileTemplateId.FIRE_BULLETS:
                break
            if not _fire_bullets_active(
                players,
                state=state,
                owner=owner,
                owner_player_index=owner_player_index,
            ):
                break
            type_id = ProjectileTemplateId.FIRE_BULLETS

    meta = travel_budget_for_type_id(type_id)
    return state.projectiles.spawn(
        pos=pos,
        angle=float(angle),
        type_id=type_id,
        owner=owner,
        travel_budget=float(meta),
        hits_players=bool(hits_players),
    )


def spawn_projectile_ring(
    state: GameplayState,
    origin_pos: Vec2,
    *,
    count: int,
    angle_offset: float,
    type_id: ProjectileTemplateId,
    owner: OwnerRef,
    owner_player_index: int | None = None,
    players: list[PlayerState] | None = None,
) -> None:
    if count <= 0:
        return
    step = math.tau / float(count)
    for idx in range(count):
        projectile_spawn(
            state,
            players=players,
            pos=origin_pos,
            angle=float(idx) * step + float(angle_offset),
            type_id=type_id,
            owner=owner,
            owner_player_index=owner_player_index,
        )



from __future__ import annotations

from .presentation import PresentationLayer
from .runtime import WorldRuntime
from .sim_world_state import SimWorldState

__all__ = ["PresentationLayer", "SimWorldState", "WorldRuntime"]



from __future__ import annotations

import random
from pathlib import Path

from grim import music as grim_music
from grim.assets import PaqTextureCache, TextureLoader
from grim.audio import AudioState, init_audio_state, play_music, shutdown_audio, update_audio
from grim.config import CrimsonConfig
from grim.console import ConsoleState
from grim.fonts.grim_mono import GrimMonoFont, load_grim_mono_font
from grim.fonts.small import SmallFontData, draw_small_text, load_small_font, measure_small_text_width
from grim.geom import Vec2
from grim.raylib_api import rl
from grim.view import ViewContext

from ..game_modes import GameMode
from ..render.rtx.mode import mode_from_rtx_flag
from ..replay import (
    Replay,
    apply_replay_bootstrap,
    load_replay_file,
    warn_on_game_version_mismatch,
)
from ..replay.types import ReplayHeader
from ..sim.batch_apply import (
    PresentationTickOutput,
    apply_presentation_outputs,
)
from ..sim.bootstrap import BOOTSTRAP_KIND_TERRAIN_V1
from ..sim.clock import FixedStepClock
from ..sim.driver.playback_driver import (
    PlaybackDriver,
    build_runtime_playback_driver,
    resolve_replay_quest_setup,
)
from ..sim.driver.playback_pump import advance_playback_frame
from ..sim.driver.setup import ReplayRunnerError, status_from_snapshot
from ..sim.hooks import TickResult
from ..sim.presentation_reactions import (
    PostApplyReaction,
    apply_post_apply_reaction,
    build_post_apply_reaction,
)
from ..terrain_assets import terrain_texture_by_id
from ..ui.hud import (
    HUD_AMMO_BASE_POS,
    HUD_AMMO_TEXT_OFFSET,
    HudAssets,
    HudRenderContext,
    HudState,
    draw_hud_overlay,
    hud_flags_for_game_mode,
    hud_ui_scale,
    load_hud_assets,
)
from ..views.quest_run_overlay import (
    draw_quest_complete_banner_overlay,
    draw_quest_title_timer_overlay,
    quest_level_label,
)
from ..weapons import WeaponId
from ..world.runtime import WorldRuntime
from ..world.terrain_runtime import normalize_terrain_ids

_PLAYBACK_SPEED_STEPS: tuple[float, ...] = (0.25, 0.5, 1.0, 2.0, 4.0, 8.0)
_DEFAULT_SPEED_INDEX = 2
_SKIP_SHORT_SECONDS = 5.0
_SKIP_LONG_SECONDS = 30.0
_REPLAY_WIDGET_PANEL_SIZE = Vec2(182.0, 53.0)
_REPLAY_WIDGET_ICON_SIZE = Vec2(32.0, 32.0)
_REPLAY_WIDGET_BAR_HEIGHT = 4.0
_REPLAY_WIDGET_X_SHIFT = 10.0
_REPLAY_WIDGET_TEXT_LINE1_Y = HUD_AMMO_BASE_POS[1] + HUD_AMMO_TEXT_OFFSET[1]
_REPLAY_WIDGET_PANEL_TO_LINE1_Y = -7.0
_REPLAY_WIDGET_PANEL_OFFSET_X = 0.0
_REPLAY_WIDGET_PANEL_OFFSET_Y = 0.0
_REPLAY_WIDGET_CLOCK_OFFSET_X = 0.0
_REPLAY_WIDGET_CLOCK_OFFSET_Y = 0.0
_REPLAY_WIDGET_TEXT_OFFSET_X = 0.0
_REPLAY_WIDGET_TEXT_OFFSET_Y = 0.0
_REPLAY_WIDGET_BAR_OFFSET_X = 0.0
_REPLAY_WIDGET_BAR_OFFSET_Y = 0.0


def _world_reset_seed_for_replay(header: ReplayHeader) -> int:
    match str(header.bootstrap_kind):
        case kind if kind == BOOTSTRAP_KIND_TERRAIN_V1:
            return int(header.bootstrap_seed)
        case _:
            return int(header.seed)


class ReplayPlaybackMode:
    def __init__(
        self,
        ctx: ViewContext,
        *,
        replay_path: Path,
        config: CrimsonConfig,
        console: ConsoleState,
        max_ticks: int | None = None,
        trace_rng: bool = False,
        rtx: bool = False,
        show_replay_widget: bool = True,
    ) -> None:
        self._ctx = ctx
        self._replay_path = Path(replay_path)
        self._config = config
        self._console = console
        self._max_ticks = max(0, int(max_ticks)) if max_ticks is not None else None
        self._trace_rng = bool(trace_rng)
        self._rtx = bool(rtx)
        self._show_replay_widget = bool(show_replay_widget)

        self.close_requested = False

        self._replay: Replay | None = None
        self._texture_cache: PaqTextureCache | None = None
        self._runtime: WorldRuntime | None = None
        self._small: SmallFontData | None = None
        self._hud_assets: HudAssets | None = None
        self._hud_state = HudState()
        self._grim_mono: GrimMonoFont | None = None
        self._quest_complete_texture: rl.Texture | None = None
        self._quest_title = ""
        self._quest_level = ""

        self._tick_rate = 60
        self._dt = 1.0 / 60.0
        self._dt_accum = 0.0
        self._clock = FixedStepClock(tick_rate=60)
        self._frame_index = 0
        self._tick_index = 0
        self._finished = False
        self._paused = False
        self._step_once_pending = False
        self._speed_index = _DEFAULT_SPEED_INDEX

        self._driver: PlaybackDriver | None = None
        self._quest_total_spawn_count = 0

        self._audio: AudioState | None = None
        self._audio_rng: random.Random | None = None

    @property
    def tick_index(self) -> int:
        return int(self._tick_index)

    @property
    def finished(self) -> bool:
        return bool(self._finished)

    @staticmethod
    def _format_time_text(seconds: float) -> str:
        total_seconds = max(0, int(seconds))
        minutes = total_seconds // 60
        rem_seconds = total_seconds % 60
        return f"{minutes}:{rem_seconds:02d}"

    def _replay_progress_ratio(self) -> float:
        replay = self._replay
        if replay is None:
            return 0.0
        total_ticks = len(replay.ticks)
        if total_ticks <= 0:
            return 1.0
        ratio = float(self._tick_index) / float(total_ticks)
        if ratio < 0.0:
            return 0.0
        if ratio > 1.0:
            return 1.0
        return ratio

    def _register_replay_audio_commands(self) -> None:
        console = self._console

        def cmd_snd_add_game_tune(args: list[str]) -> None:
            if len(args) != 1:
                console.log.log("snd_addGameTune <tuneName.ogg>")
                return
            audio = self._audio
            if audio is None:
                return
            rel_path = f"music/{args[0]}"
            result = grim_music.load_music_track(audio.music, self._ctx.assets_dir, rel_path, console=console)
            if result is None:
                return
            track_key, _track_id = result
            grim_music.queue_track(audio.music, track_key)

        console.register_command("snd_addGameTune", cmd_snd_add_game_tune)

    def _load_game_tune_queue(self) -> None:
        if self._audio is None:
            return
        self._console.exec_line("exec music/game_tunes.txt")

    def _draw_world(self, *, draw_aim_indicators: bool = True, entity_alpha: float = 1.0) -> None:
        runtime = self._runtime
        if runtime is None:
            return
        runtime.render_resources.bake_fx_queues()
        runtime.renderer.draw(
            render_frame=runtime.build_render_frame(),
            draw_aim_indicators=draw_aim_indicators,
            entity_alpha=entity_alpha,
        )

    def _replay_widget_metrics(self) -> tuple[float, float, float, float, float, float]:
        screen_w = float(rl.get_screen_width())
        screen_h = float(rl.get_screen_height())
        scale = hud_ui_scale(screen_w, screen_h)

        panel_w = _REPLAY_WIDGET_PANEL_SIZE.x * scale
        panel_h = _REPLAY_WIDGET_PANEL_SIZE.y * scale
        panel_x = screen_w - panel_w - float(_REPLAY_WIDGET_X_SHIFT) * scale
        line1_y = float(_REPLAY_WIDGET_TEXT_LINE1_Y) * scale
        panel_y = max(2.0 * scale, line1_y + _REPLAY_WIDGET_PANEL_TO_LINE1_Y * scale)
        return scale, panel_x, panel_y, panel_w, panel_h, line1_y

    def _draw_replay_widget(self) -> None:
        replay = self._replay
        if replay is None:
            return

        scale, panel_x, panel_y, panel_w, panel_h, line1_y = self._replay_widget_metrics()
        panel_x += float(_REPLAY_WIDGET_PANEL_OFFSET_X) * scale
        panel_y += float(_REPLAY_WIDGET_PANEL_OFFSET_Y) * scale

        assets = self._hud_assets

        icon_w = _REPLAY_WIDGET_ICON_SIZE.x * scale
        icon_h = _REPLAY_WIDGET_ICON_SIZE.y * scale
        icon_x = panel_x + 2.0 * scale + float(_REPLAY_WIDGET_CLOCK_OFFSET_X) * scale
        icon_y = panel_y + 8.0 * scale + float(_REPLAY_WIDGET_CLOCK_OFFSET_Y) * scale

        if assets is not None and assets.clock_table is not None:
            src = rl.Rectangle(0.0, 0.0, float(assets.clock_table.width), float(assets.clock_table.height))
            dst = rl.Rectangle(icon_x, icon_y, icon_w, icon_h)
            rl.draw_texture_pro(assets.clock_table, src, dst, rl.Vector2(0.0, 0.0), 0.0, rl.Color(255, 255, 255, 230))

        elapsed_seconds = float(self._tick_index) / float(self._tick_rate)

        if assets is not None and assets.clock_pointer is not None:
            src = rl.Rectangle(0.0, 0.0, float(assets.clock_pointer.width), float(assets.clock_pointer.height))
            center_x = icon_x + icon_w * 0.5
            center_y = icon_y + icon_h * 0.5
            dst = rl.Rectangle(center_x, center_y, icon_w, icon_h)
            origin = rl.Vector2(icon_w * 0.5, icon_h * 0.5)
            rotation = max(0.0, float(elapsed_seconds)) * 6.0
            rl.draw_texture_pro(
                assets.clock_pointer,
                src,
                dst,
                origin,
                rotation,
                rl.Color(255, 255, 255, 220),
            )

        total_ticks = len(replay.ticks)
        total_seconds = float(total_ticks) / float(self._tick_rate)
        progress_ratio = self._replay_progress_ratio()

        text_x = icon_x + icon_w + 6.0 * scale + float(_REPLAY_WIDGET_TEXT_OFFSET_X) * scale
        line1_y = line1_y + float(_REPLAY_WIDGET_TEXT_OFFSET_Y) * scale
        text_scale = 1.0
        status = "PAUSE" if self._paused else "REPLAY"
        status_color = rl.Color(245, 210, 120, 230) if self._paused else rl.Color(230, 230, 230, 220)
        self._draw_ui_text(
            f"{status} {self._playback_speed():.2f}x",
            Vec2(text_x, line1_y),
            status_color,
            scale=text_scale,
        )

        elapsed_text = self._format_time_text(elapsed_seconds)
        total_text = self._format_time_text(total_seconds)
        elapsed_w = self._measure_ui_text_width(elapsed_text, scale=text_scale)
        total_w = self._measure_ui_text_width(total_text, scale=text_scale)
        line2_y = line1_y + 18.0 * scale

        right_limit = panel_x + panel_w - 4.0 * scale + float(_REPLAY_WIDGET_TEXT_OFFSET_X) * scale
        total_x = right_limit - total_w
        bar_x_base = text_x + elapsed_w + 6.0 * scale
        bar_w = max(8.0 * scale, total_x - 6.0 * scale - bar_x_base)
        bar_x = bar_x_base + float(_REPLAY_WIDGET_BAR_OFFSET_X) * scale
        bar_y = line2_y + 5.0 * scale + float(_REPLAY_WIDGET_BAR_OFFSET_Y) * scale
        bar_h = _REPLAY_WIDGET_BAR_HEIGHT * scale
        rl.draw_rectangle(int(bar_x), int(bar_y), int(bar_w), int(bar_h), rl.Color(46, 67, 96, 150))
        fill_w = bar_w * progress_ratio
        if fill_w > 0.0:
            rl.draw_rectangle(int(bar_x), int(bar_y), int(fill_w), int(bar_h), rl.Color(70, 130, 220, 225))

        self._draw_ui_text(
            elapsed_text,
            Vec2(text_x, line2_y),
            rl.Color(220, 220, 220, 210),
            scale=text_scale,
        )
        self._draw_ui_text(
            total_text,
            Vec2(total_x, line2_y),
            rl.Color(220, 220, 220, 210),
            scale=text_scale,
        )

    def open(self) -> None:
        self._small = load_small_font(self._ctx.assets_dir)
        self._hud_assets = load_hud_assets(self._ctx.assets_dir)
        self._hud_state = HudState()
        if self._grim_mono is not None:
            rl.unload_texture(self._grim_mono.texture)
        self._grim_mono = None
        self._quest_complete_texture = None
        self._quest_title = ""
        self._quest_level = ""

        replay = load_replay_file(self._replay_path)
        self._replay = replay
        warn_on_game_version_mismatch(replay, action="playback")

        tick_rate = int(replay.header.tick_rate)
        if tick_rate <= 0:
            raise ValueError(f"invalid tick_rate: {tick_rate}")
        self._tick_rate = tick_rate
        self._dt = 1.0 / float(tick_rate)
        self._dt_accum = 0.0
        self._clock = FixedStepClock(tick_rate=int(tick_rate))
        self._frame_index = 0
        self._tick_index = 0
        self._finished = False
        self._paused = False
        self._step_once_pending = False
        self._speed_index = _DEFAULT_SPEED_INDEX
        self._driver = None

        world_size = float(replay.header.world_size)
        audio = init_audio_state(self._config, self._ctx.assets_dir, self._console)
        audio_rng = random.Random(int(replay.header.seed) & 0xFFFFFFFF)
        self._audio = audio
        self._audio_rng = audio_rng
        self._register_replay_audio_commands()
        self._load_game_tune_queue()

        quest_fail_retry_count = int(replay.header.quest_fail_retry_count)
        hardcore = bool(replay.header.hardcore)
        preserve_bugs = bool(replay.header.preserve_bugs)
        rtx_mode = mode_from_rtx_flag(self._rtx)

        runtime = WorldRuntime(
            assets_dir=self._ctx.assets_dir,
            world_size=float(world_size),
            demo_mode_active=False,
            quest_fail_retry_count=int(quest_fail_retry_count),
            hardcore=bool(hardcore),
            preserve_bugs=bool(preserve_bugs),
            texture_cache=self._texture_cache,
            config=self._config,
            audio=self._audio,
            audio_rng=self._audio_rng,
            rtx_mode=rtx_mode,
        )
        self._runtime = runtime
        runtime.reset(
            seed=_world_reset_seed_for_replay(replay.header),
            player_count=int(replay.header.player_count),
        )
        runtime.open_runtime()
        self._texture_cache = runtime.texture_cache

        sim_world = runtime.sim_world
        render_resources = runtime.render_resources

        self._hud_state.preserve_bugs = bool(sim_world.state.preserve_bugs)
        sim_world.state.status = status_from_snapshot(
            quest_unlock_index=int(replay.header.status.quest_unlock_index),
            quest_unlock_index_full=int(replay.header.status.quest_unlock_index_full),
            weapon_usage_counts=replay.header.status.weapon_usage_counts,
        )
        bootstrap = apply_replay_bootstrap(
            replay.header,
            rng=sim_world.state.rng,
            world_size=float(world_size),
        )
        if bootstrap is not None:
            runtime.terrain_runtime.apply_bootstrap_terrain(
                terrain_ids=bootstrap.terrain.terrain_ids,
                seed=int(bootstrap.terrain.terrain_seed),
                layers=3,
            )

        mode_id = replay.header.game_mode_id
        spawn_entries = None
        quest_stage_major: int | None = None
        quest_stage_minor: int | None = None
        start_weapon_id: WeaponId | None = None
        match mode_id:
            case GameMode.QUESTS:
                quest, spawn_entries = resolve_replay_quest_setup(
                    replay,
                    world_size=float(world_size),
                    player_count=len(sim_world.players),
                )

                self._quest_title = str(quest.title)
                self._quest_level = quest_level_label(quest.major, quest.minor)
                self._grim_mono = load_grim_mono_font(self._ctx.assets_dir)
                self._quest_complete_texture = self._load_quest_complete_texture()
                quest_stage_major, quest_stage_minor = quest.level_key

                base_id, overlay_id, detail_id = normalize_terrain_ids(quest.terrain_ids)
                base = terrain_texture_by_id(base_id)
                overlay = terrain_texture_by_id(overlay_id)
                detail = terrain_texture_by_id(detail_id)
                if base is not None and overlay is not None:
                    base_key, base_path = base
                    overlay_key, overlay_path = overlay
                    detail_key = detail[0] if detail is not None else None
                    detail_path = detail[1] if detail is not None else None
                    runtime.terrain_runtime.set_terrain(
                        base_key=base_key,
                        overlay_key=overlay_key,
                        base_path=base_path,
                        overlay_path=overlay_path,
                        detail_key=detail_key,
                        detail_path=detail_path,
                    )
                    runtime.terrain_runtime.schedule_from_rng_seed(
                        seed=int(sim_world.state.rng.state),
                        layers=3,
                    )

                start_weapon_id = quest.start_weapon_id
                if start_weapon_id <= WeaponId.NONE:
                    start_weapon_id = WeaponId.PISTOL
                self._quest_total_spawn_count = int(sum(int(entry.count) for entry in spawn_entries))
            case _:
                self._quest_total_spawn_count = 0

        try:
            self._driver = build_runtime_playback_driver(
                replay,
                max_ticks=self._max_ticks,
                trace_rng=bool(self._trace_rng),
                world_size=float(world_size),
                fx_queue=render_resources.fx_queue,
                fx_queue_rotated=render_resources.fx_queue_rotated,
                spawn_entries=spawn_entries,
                quest_stage_major=quest_stage_major,
                quest_stage_minor=quest_stage_minor,
                start_weapon_id=start_weapon_id,
            )
            sim_world.load_world_state(self._driver.world)
        except ReplayRunnerError as exc:  # pragma: no cover
            raise ValueError(f"unsupported replay game_mode_id: {int(mode_id)}") from exc

    def close(self) -> None:
        if self._small is not None:
            rl.unload_texture(self._small.texture)
            self._small = None
        if self._grim_mono is not None:
            rl.unload_texture(self._grim_mono.texture)
            self._grim_mono = None
        self._quest_complete_texture = None
        self._hud_assets = None
        self._driver = None
        if self._runtime is not None:
            self._runtime.close_runtime()
            self._runtime = None
        if self._audio is not None:
            shutdown_audio(self._audio)
            self._audio = None
            self._audio_rng = None

    def should_close(self) -> bool:
        return bool(self.close_requested)

    def consume_screenshot_request(self) -> bool:
        return False

    def _draw_ui_text(self, text: str, pos: Vec2, color: rl.Color, *, scale: float = 1.0) -> None:
        if self._small is not None:
            draw_small_text(self._small, text, pos, scale, color)
        else:
            rl.draw_text(text, int(pos.x), int(pos.y), int(20 * scale), color)

    def _measure_ui_text_width(self, text: str, *, scale: float = 1.0) -> float:
        if self._small is not None:
            return float(measure_small_text_width(self._small, text, scale))
        return float(len(text)) * 8.0 * float(scale)

    def _load_quest_complete_texture(self) -> rl.Texture | None:
        loader = TextureLoader(
            assets_root=self._ctx.assets_dir,
            cache=self._texture_cache,
        )
        return loader.get(
            name="ui_textLevComp",
            paq_rel="ui/ui_textLevComp.jaz",
        )

    def _build_post_apply_reaction(self, *, tick_result: TickResult) -> PostApplyReaction:
        driver = self._driver
        return build_post_apply_reaction(
            tick_result=tick_result,
            quest_state=(
                None
                if driver is None
                else driver.quest_spawn_state
            ),
        )

    def _apply_post_apply_reaction(self, reaction: PostApplyReaction) -> None:
        runtime = self._runtime
        if runtime is None:
            return
        apply_post_apply_reaction(
            reaction=reaction,
            play_sfx=runtime.audio_bridge.router.play_sfx,
            play_completion_music=self._play_quest_completion_music,
        )

    def _play_quest_completion_music(self) -> None:
        if self._audio is None:
            return
        play_music(self._audio, "crimsonquest")
        playback = self._audio.music.playbacks.get("crimsonquest")
        if playback is None:
            return
        playback.volume = 0.0
        try:
            rl.set_music_volume(playback.music, 0.0)
        except RuntimeError:
            playback.volume = 0.0

    def _tick_limit(self) -> int:
        replay = self._replay
        if replay is None:
            return 0
        total_ticks = len(replay.ticks)
        if self._max_ticks is None:
            return int(total_ticks)
        return min(int(total_ticks), max(0, int(self._max_ticks)))

    def _mark_finished_if_complete(self) -> None:
        tick_limit = int(self._tick_limit())
        if int(self._tick_index) < int(tick_limit):
            return
        self._finished = True

    def _advance_runner(
        self,
        *,
        dt_seconds: float,
        max_ticks: int | None = None,
        bake_fx_per_tick: bool = False,
    ) -> None:
        replay = self._replay
        runtime = self._runtime
        driver = self._driver
        if replay is None or runtime is None or driver is None:
            self._finished = True
            return
        render_resources = runtime.render_resources
        tick_limit = int(self._tick_limit())
        if int(self._tick_index) >= tick_limit:
            self._mark_finished_if_complete()
            return

        frame_dt = float(dt_seconds)
        advance = advance_playback_frame(
            driver=driver,
            sim_world=runtime.sim_world,
            clock=self._clock,
            start_tick=int(self._tick_index),
            frame_index=int(self._frame_index),
            dt_seconds=float(frame_dt),
            max_ticks=max_ticks,
            tick_limit=int(tick_limit),
            game_tune_started=bool(driver.session.game_tune_started),
        )
        self._frame_index = int(advance.frame_index)
        self._tick_index = int(advance.next_tick_index)

        def _on_output_applied(output: PresentationTickOutput) -> None:
            self._apply_post_apply_reaction(reaction_by_tick[int(output.tick_index)])
            if not bake_fx_per_tick:
                return
            if render_resources.ground is not None and render_resources.fx_textures is not None:
                render_resources.bake_fx_queues()
            else:
                render_resources.fx_queue.clear()
                render_resources.fx_queue_rotated.clear()

        reaction_by_tick = {
            int(tick_result.source_tick.tick_index): self._build_post_apply_reaction(tick_result=tick_result)
            for tick_result in advance.tick_results
        }
        if advance.outputs:
            apply_presentation_outputs(
                outputs=advance.outputs,
                sync_audio_bridge_state=runtime.sync_audio_bridge_state,
                apply_audio_plan=lambda plan, should_apply_audio: runtime.audio_bridge.apply_plan(
                    plan=plan,
                    apply_audio=bool(should_apply_audio),
                ),
                update_camera=runtime.update_camera,
                on_output_applied=_on_output_applied,
                apply_audio=True,
            )
        else:
            for output in advance.outputs:
                _on_output_applied(output)

        self._mark_finished_if_complete()
        self._dt_accum = float(self._clock.accum)

    def _playback_speed(self) -> float:
        return float(_PLAYBACK_SPEED_STEPS[int(self._speed_index)])

    def _change_speed(self, delta: int) -> None:
        idx = int(self._speed_index) + int(delta)
        idx = max(0, min(idx, len(_PLAYBACK_SPEED_STEPS) - 1))
        self._speed_index = idx

    def _skip_forward_seconds(self, seconds: float) -> None:
        replay = self._replay
        if replay is None or self._finished:
            return
        ticks = int(round(float(seconds) * float(self._tick_rate)))
        if ticks <= 0:
            return
        target = min(int(self._tick_limit()), int(self._tick_index) + int(ticks))
        audio_bridge = self._runtime.audio_bridge if self._runtime is not None else None
        prev_sfx_enabled: bool | None = None
        if audio_bridge is not None and audio_bridge.router is not None:
            prev_sfx_enabled = bool(audio_bridge.router.sfx_enabled)
            audio_bridge.router.sfx_enabled = False
        try:
            ticks_to_advance = max(0, int(target) - int(self._tick_index))
            if ticks_to_advance > 0:
                self._advance_runner(
                    dt_seconds=float(ticks_to_advance) * float(self._dt),
                    max_ticks=int(ticks_to_advance),
                    bake_fx_per_tick=True,
                )
        finally:
            if prev_sfx_enabled is not None and audio_bridge is not None and audio_bridge.router is not None:
                audio_bridge.router.sfx_enabled = bool(prev_sfx_enabled)
        self._clock.reset()
        self._dt_accum = 0.0

    def update(self, dt: float) -> None:
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ESCAPE):
            self.close_requested = True
            return
        if rl.is_key_pressed(rl.KeyboardKey.KEY_SPACE):
            self._paused = not bool(self._paused)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_PERIOD) and bool(self._paused):
            self._step_once_pending = True
        if rl.is_key_pressed(rl.KeyboardKey.KEY_LEFT_BRACKET):
            self._change_speed(-1)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT_BRACKET):
            self._change_speed(1)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_ONE):
            self._speed_index = _DEFAULT_SPEED_INDEX
        if rl.is_key_pressed(rl.KeyboardKey.KEY_RIGHT):
            self._skip_forward_seconds(_SKIP_SHORT_SECONDS)
        if rl.is_key_pressed(rl.KeyboardKey.KEY_PAGE_DOWN):
            self._skip_forward_seconds(_SKIP_LONG_SECONDS)

        if not self._finished and bool(self._paused) and bool(self._step_once_pending):
            self._clock.reset()
            self._advance_runner(
                dt_seconds=float(self._dt),
                max_ticks=1,
            )
            self._clock.reset()
            self._step_once_pending = False
            self._dt_accum = 0.0

        if not self._finished and (not self._paused):
            dt = float(dt)
            if dt < 0.0:
                dt = 0.0
            if dt > 0.1:
                dt = 0.1
            self._advance_runner(
                dt_seconds=dt * self._playback_speed(),
            )

        if self._audio is not None:
            update_audio(self._audio, float(dt))

        # Runtime open schedules terrain generation, but replay advances
        # deterministic world ticks directly, so we must process pending ground
        # work explicitly.
        if self._runtime is not None:
            ground = self._runtime.render_resources.ground
            if ground is not None:
                ground.process_pending()

    def _draw_quest_title(self) -> None:
        replay = self._replay
        if replay is None or replay.header.game_mode_id != GameMode.QUESTS:
            return
        font = self._grim_mono
        if font is None:
            return
        title = str(self._quest_title or "")
        level = str(self._quest_level or "")
        if not title or not level:
            return
        driver = self._driver
        if driver is None or driver.quest_spawn_state is None:
            return

        draw_quest_title_timer_overlay(
            font,
            title,
            level,
            timer_ms=float(driver.quest_spawn_state.spawn_timeline_ms),
        )

    def _draw_quest_complete_banner(self) -> None:
        replay = self._replay
        if replay is None or replay.header.game_mode_id != GameMode.QUESTS:
            return
        tex = self._quest_complete_texture
        if tex is None:
            return
        driver = self._driver
        if driver is None or driver.quest_spawn_state is None:
            return
        draw_quest_complete_banner_overlay(
            tex,
            timer_ms=float(driver.quest_spawn_state.completion_transition_ms),
        )

    def draw(self) -> None:
        sim_world = self._runtime.sim_world if self._runtime is not None else None
        if sim_world is not None:
            self._draw_world(draw_aim_indicators=True)
        else:
            rl.clear_background(rl.BLACK)

        replay = self._replay
        if (
            sim_world is not None
            and replay is not None
            and self._hud_assets is not None
            and sim_world.players
        ):
            mode_id = replay.header.game_mode_id
            hud_flags = hud_flags_for_game_mode(mode_id)
            quest_progress_ratio: float | None = None
            elapsed_ms = float(sim_world.presentation_elapsed_ms)
            match mode_id:
                case GameMode.QUESTS:
                    total = int(self._quest_total_spawn_count)
                    kills = int(sim_world.creatures.kill_count)
                    quest_progress_ratio = float(kills) / float(total) if total > 0 else None
                    driver = self._driver
                    if driver is not None:
                        elapsed_ms = float(driver.elapsed_ms)
                case _:
                    driver = self._driver
                    if driver is not None:
                        elapsed_ms = float(driver.elapsed_ms)
            draw_hud_overlay(
                HudRenderContext(
                    assets=self._hud_assets,
                    state=self._hud_state,
                    font=self._small,
                    show_health=bool(hud_flags.show_health),
                    show_weapon=bool(hud_flags.show_weapon),
                    show_xp=bool(hud_flags.show_xp),
                    show_time=bool(hud_flags.show_time),
                    show_quest_hud=bool(hud_flags.show_quest_hud),
                    small_indicators=False,
                ),
                player=sim_world.players[0],
                players=sim_world.players,
                bonus_hud=sim_world.state.bonus_hud,
                elapsed_ms=elapsed_ms,
                frame_dt_ms=float(max(0.0, rl.get_frame_time()) * 1000.0),
                quest_progress_ratio=quest_progress_ratio,
            )

        self._draw_quest_title()
        self._draw_quest_complete_banner()

        if bool(self._show_replay_widget):
            self._draw_replay_widget()



from __future__ import annotations

from .canonical_channels import RngStreamRow

_RNG_MARKS_EMPTY_VALUE = -1


def canonical_rng_marks(
    *,
    rng_state: int,
    rng_stream: list[RngStreamRow],
) -> dict[str, int]:
    rows = list(rng_stream)
    call_count = len(rows)

    first_value = _RNG_MARKS_EMPTY_VALUE
    last_value = _RNG_MARKS_EMPTY_VALUE
    first_state_before = _RNG_MARKS_EMPTY_VALUE
    last_state_after = _RNG_MARKS_EMPTY_VALUE
    if rows:
        first = rows[0]
        last = rows[-1]
        first_value = int(first.value_15)
        last_value = int(last.value_15)
        first_state_before = int(first.state_before_u32)
        last_state_after = int(last.state_after_u32)

    return {
        "calls_total": int(call_count),
        "first_value_15": int(first_value),
        "last_value_15": int(last_value),
        "first_state_before_u32": int(first_state_before),
        "last_state_after_u32": int(last_state_after),
        "checkpoint_rng_state": int(rng_state),
    }
